from typing import Sequence, Union

import numpy as np
from arbol import aprint, asection

from dexp.datasets.base_dataset import BaseDataset
from dexp.processing.backends.best_backend import BestBackend
from dexp.utils.slicing import slice_from_shape
from dexp.processing.remove_beads import BeadsRemover


def dataset_extract_psf(dataset: BaseDataset,
                        dest_path: str,
                        channels: Sequence[str],
                        slicing: Union[Sequence[slice], slice],
                        peak_threshold: int = 500,
                        similarity_threshold: float = 0.5,
                        psf_size: int = 35,
                        device: int = 0,
                        stop_at_exception: bool = True,
                        verbose: bool = True,
                        ) -> None:
    """
    Computes PSF from beads.
    Additional information at dexp.processing.remove_beads.beadremover documentation.
    """
    dest_path = dest_path.split('.')[0]

    remove_beads = BeadsRemover(
        peak_threshold=peak_threshold,
        similarity_threshold=similarity_threshold,
        psf_size=psf_size,
        verbose=verbose
    )

    for channel in dataset._selected_channels(channels):
        array = dataset.get_array(channel)
        _, volume_slicing, time_points = slice_from_shape(array.shape, slicing)

        psfs = []
        for i in range(len(time_points)):
            tp = time_points[i]
            try:
                with asection(f'Removing beads of channel: {channel}'):
                    with asection(f'Loading time point {i}/{len(time_points)}'):
                        tp_array = np.asarray(array[tp][volume_slicing])

                    with asection('Processing'):
                        with BestBackend(exclusive=True, enable_unified_memory=True, device_id=device) as backend:
                            tp_array = backend.to_backend(tp_array)
                            estimated_psf = remove_beads.detect_beads(tp_array)

                aprint(f"Done extracting PSF from time point: {i}/{len(time_points)} .")

            except Exception as error:
                aprint(error)
                aprint(f"Error occurred while processing time point {i} !")
                import traceback
                traceback.print_exc()

                if stop_at_exception:
                    raise error
            
            psfs.append(estimated_psf)

        psfs = np.stack(psfs).mean(axis=0)
        print(psfs.shape)
        np.save(dest_path + channel + '.npy', psfs)



from typing import Optional, List, Tuple

import numpy as np
from numpy.typing import ArrayLike
from arbol import aprint

from dexp.processing.backends.backend import Backend
from dexp.utils import xpArray


class Bead:
    def __init__(self, image: ArrayLike, coord: Optional[Tuple[int]], size: int):
        if coord is not None:
            self.slicing = self.get_slice(coord, size)
            self.data = image[self.slicing]
        else:
            self.slicing = None
            self.data = image
        self.data = Backend.current().to_numpy(self.data)
        self.data = self.data / np.linalg.norm(self.data)
    
    @property
    def shape(self) -> Tuple[int]:
        return self.data.shape

    @staticmethod
    def get_slice(coord: Tuple[int], size: int) -> List[slice]:
        slicing = tuple(slice(c - size // 2, c + size // 2 + 1) for c in coord)
        return slicing
    
    @staticmethod
    def from_data(array: ArrayLike) -> 'Bead':
        return Bead(array, None, -1)
    
    def __array__(self, *args) -> ArrayLike:
        return self.data.reshape(-1)


class BeadsRemover:
    def __init__(self, peak_threshold: int, similarity_threshold: float,
                 psf_size: int = 31, verbose: bool = False):
        """
        Helper object to detect beads, extract the PSF from it and remove it if necessary.

        Args:
            peak_threshold (int): Peak maxima threshold for object (beads) detection.
                Lower values detect more objects, but it might get more false positives.
            similarity_threshold (float): Threshold for beads selection given their cosine similarity with the estimated (median) PSF.
            psf_size (int, optional): PSF size (shape). Defaults to 31.
            verbose (bool, optional): Flag to display intermediate results. Defaults to False.
        """
        self.peak_threshold = peak_threshold
        self.similarity_threshold = similarity_threshold
        self.psf_size = psf_size
        self.beads: List[Bead] = []

        self.verbose = verbose
    
    def detect_beads(self, array: xpArray, estimated_psf: Optional[ArrayLike] = None) -> ArrayLike:
        from cucim.skimage.feature import peak_local_max

        backend = Backend.current()

        sp = backend.get_sp_module()
        xp = backend.get_xp_module()

        if estimated_psf is None:
            aprint('Finding beads with point detection.')
            kernel = xp.zeros((1, 5, 5))
            kernel[0, 2, 2] = 4
            kernel[0, 2, 0] = -1
            kernel[0, 0, 2] = -1
            kernel[0, 2, 4] = -1
            kernel[0, 4, 2] = -1
        else:
            aprint('Finding beads with estimated PSF.')
            kernel = estimated_psf.copy()
            dark_region = kernel < kernel.mean()
            bright_region = np.logical_not(dark_region)
            # equalize the values of both the regions to be 1.0
            kernel[dark_region] = - 1.0  / dark_region.sum()
            kernel[bright_region] /= kernel[bright_region].sum()

            kernel = backend.to_backend(kernel.astype(xp.float32))

        # try detecting beads with given PSF or dot detection
        filtered = sp.ndimage.correlate(array.astype(xp.float32), weights=kernel)

        # small blur to center detection
        mean_kernel = xp.ones((5,) * array.ndim, dtype=xp.float32)
        mean_kernel /= mean_kernel.sum()
        filtered = sp.ndimage.correlate(filtered, weights=mean_kernel)

        # find local maxima in filtered image
        coordinates = peak_local_max(filtered, min_distance=self.psf_size, threshold_abs=500).get()
        
        # find local maxima in original image near the filtered maxima
        peaks = xp.zeros(array.shape, dtype=bool)
        peaks[tuple(coordinates.T)] = True
        peaks = sp.ndimage.binary_dilation(peaks, xp.ones((self.psf_size,) * array.ndim, dtype=bool))
        coordinates = peak_local_max(array, min_distance=self.psf_size, labels=peaks).get()

        if self.verbose:
            import napari
            viewer = napari.Viewer()
            viewer.add_image(filtered.get(), name='filtered', blending='additive', colormap='red', contrast_limits=(0, 5000))
            viewer.add_image(array.get(), name='original', blending='additive', colormap='green', contrast_limits=(0, 5000))
            peaks = xp.zeros(array.shape, dtype=bool)
            peaks[tuple(coordinates.T)] = True
            viewer.add_image(peaks.get(), name='peaks', blending='additive', colormap='gray', contrast_limits=(0, 1))
            napari.run()
        
        array = array.get()
        beads = []
        for coord in coordinates:
            bead = Bead(array, coord, self.psf_size)
            if np.all(np.equal(bead.shape, self.psf_size)):
                beads.append(bead)

        avg_bead = Bead.from_data(
            np.median(np.stack(tuple(beads)), axis=0).reshape((self.psf_size,) * array.ndim)
        )

        # select regions that are similar to the median estimated bead
        selected_beads = []
        for bead in beads:
            angle = np.dot(bead, avg_bead)
            if angle > self.similarity_threshold:
                selected_beads.append(bead)
                array[bead.slicing] = 255

        aprint(f'Detected {len(selected_beads)} beads.')
        
        if self.verbose:
            import napari
            viewer = napari.Viewer(ndisplay=3)
            viewer.add_image(avg_bead.data, name='estimated PSF', interpolation='nearest')
            viewer.add_image(array, contrast_limits=(0, 500), name='marked beads')
            napari.run()

        self.beads = selected_beads
        return avg_bead.data
    
    def remove_beads(self, array: xpArray, fill_value: int = 0) -> xpArray:
        for bead in self.beads:
            array[bead.slicing] = fill_value
        return array


import numpy as np
from typing import Dict, Sequence
from ome_zarr.format import CurrentFormat

def default_omero_metadata(name: str, channels: Sequence[str], dtype: np.dtype) -> Dict:
    val_max = 1.0 if np.issubdtype(dtype, np.floating) else np.iinfo(dtype).max

    return {
        'id': 1,
        'name': name,
        'version': CurrentFormat().version,
        'channels': [
            {
                'active': True,
                'coefficient': 1,
                'color': 'FFFFFF',
                'family': 'linear',
                'inverted': False,
                'label': ch,
                'window': {
                    'start': 0,
                    'end': 1500,  # guessing
                    'min': 0,
                    'max': val_max
                }
            }
            for ch in channels
        ]
    }



import numpy

from dexp.optics.psf.standard_psfs import nikon16x08na
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.deconvolution.inversion_deconvolution import inversion_deconvolution
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.utils.timeit import timeit


def demo_inversion_deconvolution_numpy():
    with NumpyBackend():
        _demo_inversion_deconvolution()


def demo_inversion_deconvolution_cupy():
    try:
        with CupyBackend():
            _demo_inversion_deconvolution()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_inversion_deconvolution(length_xy=128):
    xp = Backend.get_xp_module()

    with timeit("generate data"):
        image_gt, background, image = generate_nuclei_background_data(add_noise=False,
                                                                      length_xy=length_xy,
                                                                      length_z_factor=1,
                                                                      background_stength=0,
                                                                      add_offset=False)

    psf = nikon16x08na()
    # psf = olympus20x10na()
    psf = psf.astype(dtype=numpy.float16)
    image = image.astype(dtype=numpy.float16)

    blurry = fft_convolve(image, psf)
    blurry = blurry - blurry.min()
    blurry = blurry / blurry.max()
    noisy = blurry
    noisy = noisy + 0.1 * xp.random.uniform(size=blurry.shape)

    with timeit("deconvolved"):
        deconvolved = inversion_deconvolution(noisy, psf, mode='reflect', epsilon=1e-1)  # FIXME noisy, psf, mode='reflect')

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image), name='image')
        viewer.add_image(_c(blurry), name='blurry')
        viewer.add_image(_c(psf), name='psf')
        viewer.add_image(_c(noisy), name='noisy')
        viewer.add_image(_c(deconvolved), name='deconvolved')


if __name__ == "__main__":
    demo_inversion_deconvolution_cupy()
    demo_inversion_deconvolution_numpy()



import os
from arbol import aprint


def compute_num_workers(n_workers: int, n_time_points: int) -> int:
    if n_workers < 0:
        n_workers = int(os.cpu_count() / -n_workers)
    n_workers = min(max(1, n_workers), n_time_points)
    aprint(f"Number of workers: {n_workers}")
    return n_workers



import dask
import numpy as np
from arbol.arbol import aprint
from arbol.arbol import asection
from toolz import curry

from dexp.datasets.zarr_dataset import ZDataset
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.best_backend import BestBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.multiview_lightsheet.fusion.mvsols import msols_fuse_1C2L
from dexp.processing.multiview_lightsheet.fusion.simview import SimViewFusion
from dexp.processing.registration.model.model_io import model_list_from_file, model_list_to_file
from dexp.utils.slicing import slice_from_shape


def dataset_fuse(dataset,
                 output_path,
                 channels,
                 slicing,
                 store,
                 compression,
                 compression_level,
                 overwrite,
                 microscope,
                 equalise,
                 equalise_mode,
                 zero_level,
                 clip_too_high,
                 fusion,
                 fusion_bias_strength_i,
                 fusion_bias_strength_d,
                 dehaze_size,
                 dark_denoise_threshold,
                 z_pad_apodise,
                 loadreg,
                 model_list_filename,
                 warpreg_num_iterations,
                 min_confidence,
                 max_change,
                 registration_edge_filter,
                 maxproj,
                 huge_dataset,
                 workers,
                 workersbackend,
                 devices,
                 check,
                 pad,
                 stop_at_exception=True):

    views = {
        channel.split('-')[-1]: dataset.get_array(channel, per_z_slice=False)
        for channel in channels
    }

    with asection(f"Views:"):
        for channel, view in views.items():
            aprint(f"View: {channel} of shape: {view.shape} and dtype: {view.dtype}")

    key = list(views.keys())[0]
    dtype = views[key].dtype
    in_nb_time_points = views[key].shape[0]
    aprint(f"Slicing with: {slicing}")
    _, volume_slicing, time_points = slice_from_shape(views[key].shape, slicing)

    if microscope == 'simview':
        views = SimViewFusion.validate_views(views)
 
    # load registration models:
    with NumpyBackend():
        if loadreg:
            aprint(f"Loading registration shifts from existing file! ({model_list_filename})")
            models = model_list_from_file(model_list_filename)
            if len(models) == 1:
                models = [models[0] for _ in range(in_nb_time_points)]
            elif len(models) != in_nb_time_points:
                raise ValueError(f'Number of registration models provided ({len(models)})'
                                 f'differs from number of input time points ({in_nb_time_points})')
        else:
            models = [None] * len(time_points)

    @curry
    def process(i, params, device_id=0):
        equalisation_ratios_reference, model, dest_dataset = params
        tp = time_points[i]
        try:
            with asection(f'Fusing time point for time point {i}/{len(time_points)}'):
                with asection(f"Loading channels {channels}"):
                    views_tp = {k: np.asarray(view[tp][volume_slicing]) for k, view in views.items()}

                with BestBackend(exclusive=True, enable_unified_memory=True, device_id=device_id):
                    if models[i] is not None:
                        model = models[i]
                    # otherwise it could be a None model or from the first iteration if equalisation mode was 'first'
                    if microscope == 'simview':
                        fuse_obj = SimViewFusion(registration_model=model,
                                                 equalise=equalise,
                                                 equalisation_ratios=equalisation_ratios_reference,
                                                 zero_level=zero_level,
                                                 clip_too_high=clip_too_high,
                                                 fusion=fusion,
                                                 fusion_bias_exponent=2,
                                                 fusion_bias_strength_i=fusion_bias_strength_i,
                                                 fusion_bias_strength_d=fusion_bias_strength_d,
                                                 dehaze_before_fusion=True,
                                                 dehaze_size=dehaze_size,
                                                 dehaze_correct_max_level=True,
                                                 dark_denoise_threshold=dark_denoise_threshold,
                                                 dark_denoise_size=9,
                                                 butterworth_filter_cutoff=1,
                                                 flip_camera1=True)

                        tp_array = fuse_obj(**views_tp, pad=pad)
                        new_equalisation_ratios = fuse_obj._equalisation_ratios

                    elif microscope == 'mvsols':
                        metadata = dataset.get_metadata()
                        angle = metadata['angle']
                        channel = metadata['channel']
                        dz = metadata['dz']
                        res = metadata['res']
                        illumination_correction_sigma = metadata['ic_sigma'] if 'ic_sigma' in metadata else None

                        tp_array, model, new_equalisation_ratios = msols_fuse_1C2L(*list(views_tp.values()),
                                                                                   z_pad=z_pad_apodise[0],
                                                                                   z_apodise=z_pad_apodise[1],
                                                                                   registration_num_iterations=warpreg_num_iterations,
                                                                                   registration_force_model=loadreg,
                                                                                   registration_model=model,
                                                                                   registration_min_confidence=min_confidence,
                                                                                   registration_max_change=max_change,
                                                                                   registration_edge_filter=registration_edge_filter,
                                                                                   equalise=equalise,
                                                                                   equalisation_ratios=equalisation_ratios_reference,
                                                                                   zero_level=zero_level,
                                                                                   clip_too_high=clip_too_high,
                                                                                   fusion=fusion,
                                                                                   fusion_bias_exponent=2,
                                                                                   fusion_bias_strength_x=fusion_bias_strength_i,
                                                                                   dehaze_size=dehaze_size,
                                                                                   dark_denoise_threshold=dark_denoise_threshold,
                                                                                   angle=angle,
                                                                                   dx=res,
                                                                                   dz=dz,
                                                                                   illumination_correction_sigma=illumination_correction_sigma,
                                                                                   registration_mode='projection' if maxproj else 'full',
                                                                                   huge_dataset_mode=huge_dataset)
                    else:
                        raise NotImplementedError

                    with asection(f"Moving array from backend to numpy."):
                        tp_array = Backend.to_numpy(tp_array, dtype=dtype, force_copy=False)

                    del views_tp
                    Backend.current().clear_memory_pool()

                    aprint(f"Last equalisation ratios: {new_equalisation_ratios}")

                    # moving to numpy to allow pickling
                    if model is not None:
                        model = model.to_numpy()
                    new_equalisation_ratios = [
                        None if v is None else Backend.to_numpy(v)
                        for v in new_equalisation_ratios
                    ]

                with asection(f"Saving fused stack for time point {i}, shape:{tp_array.shape}, dtype:{tp_array.dtype}"):

                    if i == 0:
                        # We allocate last minute once we know the shape... because we don't always know the shape in advance!!!
                        mode = 'w' + ('' if overwrite else '-')
                        dest_dataset = ZDataset(output_path, mode, store, parent=dataset)
                        dest_dataset.add_channel('fused',
                                                 shape=(len(time_points),)+tp_array.shape,
                                                 dtype=tp_array.dtype,
                                                 codec=compression,
                                                 clevel=compression_level)

                    dest_dataset.write_stack(channel='fused',
                                             time_point=i,
                                             stack_array=tp_array)

                aprint(f"Done processing time point: {i}/{len(time_points)} .")

        except Exception as error:
            aprint(error)
            aprint(f"Error occurred while processing time point {i} !")
            import traceback
            traceback.print_exc()

            if stop_at_exception:
                raise error

        return new_equalisation_ratios, model, dest_dataset

    if len(devices) == 1:
        process = process(device_id=devices[0])
    else:
        from dask.distributed import Client
        from dask_cuda import LocalCUDACluster

        process = dask.delayed(process)

        cluster = LocalCUDACluster(CUDA_VISIBLE_DEVICES=devices)
        client = Client(cluster)

        aprint('Dask Client', client)

    # the parameters are (equalisation rations, registration model, dest. dataset)
    if microscope == 'simview':
        init_params = ([None, None, None], None, None)
    elif microscope == 'mvsols':
        init_params = ([None], None, None)
    else:
        raise NotImplementedError

    # it creates the output dataset from the first time point output shape
    params = process(0, init_params)
    if len(devices) > 1:
        params = params.persist()

    if equalise_mode == 'all':
        params = init_params[:2] + (params[2],)

    lazy_computations = []  # it is only lazy if len(devices) > 1
    for i in range(1, len(time_points)):
        lazy_computations.append(process(i, params))

    if len(devices) > 1:
        first_model, dest_dataset = params.compute()[1:]
        models = [first_model] + [output[1]for output in dask.compute(*lazy_computations)]
    else:
        models = [params[1]] + [output[1] for output in lazy_computations]
        dest_dataset = params[2]

    if not loadreg and models[0] is not None:
        model_list_to_file(model_list_filename, models)

    aprint(dest_dataset.info())
    if check:
        dest_dataset.check_integrity()

    # close destination dataset:
    dest_dataset.close()
    if len(devices) > 1:
        client.close()




from typing import Optional

import numpy
import scipy
from functools import reduce

from dexp.utils import xpArray
from dexp.processing.backends.backend import Backend
from dexp.processing.deconvolution.admm_utils import *


def admm_deconvolution(image: xpArray,
                       psf: xpArray,
                       rho: float = 0.1,
                       gamma: float = 0.01,
                       iterations: int = 10,
                       derivative: int = 1,
                       internal_dtype: Optional[numpy.dtype] = None,
                       display: bool = False,
                       ) -> xpArray:

    """
    Reference from: http://jamesgregson.ca/tag/admm.html
    """
    if derivative == 1:
        derivative_func = first_derivative_func
        derivative_kernels = first_derivative_kernels
    elif derivative == 2:
        derivative_func = second_derivative_func
        derivative_kernels = second_derivative_kernels
    else:
        raise RuntimeError(f'Derivative must be 1 or 2. Found {derivative}.')

    backend = Backend.current()

    if display:
        import napari
        viewer = napari.view_image(backend.to_numpy(image), name='original')

    def shrink(array: xpArray) -> xpArray:
        return xp.sign(array) * xp.clip(xp.abs(array) - gamma / rho, 0.0, None)

    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if image.ndim != psf.ndim:
        raise ValueError("The image and PSF must have same number of dimensions!")

    if internal_dtype is None:
        internal_dtype = numpy.float32

    original_dtype = image.dtype
    image = Backend.to_backend(image, dtype=internal_dtype)
    psf = Backend.to_backend(psf, dtype=internal_dtype)

    # inverting psf
    backproj = xp.flip(psf)

    # padding with reflection for a better deconvolution 
    original_shape = image.shape
    pad_width = [(s // 2, s // 2) for s in backproj.shape]
    image = xp.pad(image, pad_width, mode='reflect')
    # original_slice = tuple(slice(p, p + s) for s, (p, _) in zip(original_shape, pad_width))
    original_slice = tuple(
        slice(p - 1, p + s - 1) for s, p in zip(original_shape, backproj.shape)
    )  # I don't understand why the other slicing is wrong \O/

    # compute data shape for faster fft
    fsize = tuple(scipy.fftpack.next_fast_len(x) for x in image.shape)

    # create derivative kernels
    Ds = [
        backend.to_backend(D)
        for D in derivative_kernels(image.ndim)
    ]

    # convert derivative kernels to freq space
    fDs = [sp.fft.fftn(D, fsize) for D in Ds]
    del Ds

    # convert input to freq space
    fbackproj = sp.fft.fftn(backproj, fsize, overwrite_x=True)
    fimage = sp.fft.fftn(image, fsize)

    # pre-compute auxiliary values following reference
    nume_aux = fbackproj * fimage
    denom = fbackproj * fbackproj.conj() +\
        rho * reduce(xp.add, (fD.conj() * fD for fD in fDs))
    del fDs

    # compute parameters used for finite difference differenciation (fast diff operator)
    Daxes = derivative_axes(image.ndim)

    zeros = lambda : xp.zeros(fsize, dtype=internal_dtype)

    # allocate buffers
    I = zeros()  # output image

    Zs = [zeros() for _ in Daxes]
    Us = [zeros() for _ in Daxes]

    for _ in range(iterations):
        # iterations according to reference
        # loop operations are done using generators to reduce memory allocation
        V = rho * reduce(xp.add, (
            derivative_func(Z - U, axes, True)
            for axes, Z, U in zip(Daxes, Zs, Us)
        ))

        fV = sp.fft.fftn(V, overwrite_x=True); del V
        I = sp.fft.ifftn((nume_aux + fV) / denom, overwrite_x=True).real

        if display:
            viewer.add_image(backend.to_numpy(I[original_slice]))

        tmps = [derivative_func(I, axes, False) for axes in Daxes]
        Zs = [shrink(tmp + U) for tmp, U in zip(tmps, Us)]

        for tmp, Z, U in zip(tmps, Zs, Us):
            U += tmp - Z

    # normalization
    I = I[original_slice].astype(original_dtype)
    I -= I.min()
    I /= I.max()

    if display:
        napari.run()

    return I



import os
import re
from fnmatch import fnmatch
from os import listdir
from os.path import join, exists
from typing import Sequence, Tuple, Any

import numpy
from arbol.arbol import aprint
from cachey import Cache
from dask import array, delayed
from numpy import uint16

from dexp.datasets.base_dataset import BaseDataset
from dexp.io.compress_array import decompress_array
from dexp.utils.config import config_blosc


class CCDataset(BaseDataset):

    def __init__(self, path, cache_size=8e9):

        super().__init__(dask_backed=False)

        config_blosc()

        self.folder = path
        self._channels = []
        self._index_files = {}

        all_files = list(listdir(path))
        # print(all_files)

        for file in all_files:
            if fnmatch(file, '*.index.txt'):
                if not file.startswith('._'):
                    channel = file.replace('.index.txt', '')
                    self._channels.append(channel)
                    self._index_files[channel] = join(path, file)

        # print(self._channels)
        # print(self._index_files)

        self._nb_time_points = {}
        self._times_sec = {}
        self._shapes = {}
        self._channel_shape = {}
        self._time_points = {}

        self._is_compressed = self._find_out_if_compressed(path)

        for channel in self._channels:
            self._parse_channel(channel)

        self.cache = Cache(cache_size)  # Leverage two gigabytes of memory

    def _parse_channel(self, channel):

        index_file = self._index_files[channel]

        with open(index_file) as f:
            lines = f.readlines()

        lines = [line.strip() for line in lines]

        lines = [re.split(r'\t+', line) for line in lines]

        self._time_points[channel] = []

        for line in lines:
            time_point = int(line[0])
            time_sec = float(line[1])
            shape = eval('(' + line[2] + ')')[::-1]

            self._times_sec[(channel, time_point)] = time_sec
            self._shapes[(channel, time_point)] = shape

            if channel in self._channel_shape:
                existing_shape = self._channel_shape[channel]
                if shape != existing_shape:
                    aprint(f"Warning: Channel {channel} has varying stack shape! Shape changes from {existing_shape} to {shape} at time point {time_point}")
            self._channel_shape[channel] = shape

            self._time_points[channel].append(time_point)

        self._nb_time_points[channel] = len(self._time_points[channel])

    def _get_stack_file_name(self, channel, time_point):

        compressed_file_name = join(self.folder, 'stacks', channel, str(time_point).zfill(6) + '.blc')
        raw_file_name = join(self.folder, 'stacks', channel, str(time_point).zfill(6) + '.raw')

        if self._is_compressed is None:
            if exists(compressed_file_name):
                self._is_compressed = True
            else:
                self._is_compressed = False

        if self._is_compressed:
            return compressed_file_name
        else:
            return raw_file_name


    def _get_array_for_stack_file(self, file_name, shape=None, dtype=None):

        try:
            if file_name.endswith('.raw'):
                aprint(f"Accessing file: {file_name}")

                dt = numpy.dtype(uint16)
                dt = dt.newbyteorder('L')

                array = numpy.fromfile(file_name, dtype=dt)

            elif file_name.endswith('.blc'):
                array = numpy.empty(shape=shape, dtype=dtype)
                with open(file_name, "rb") as binary_file:
                    # Read the whole file at once
                    data = binary_file.read()
                    decompress_array(data, array)

            # Reshape array:
            if not shape is None:
                array = array.reshape(shape)

            return array


        except FileNotFoundError:
            aprint(f"Could not find file: {file_name} for array of shape: {shape}")
            return numpy.zeros(shape, dtype=uint16)

    def _get_slice_array_for_stack_file_and_z(self, file_name, shape, z):

        try:
            if file_name.endswith('.raw'):
                aprint(f"Accessing file: {file_name} at z={z}")

                length = shape[1] * shape[2] * numpy.dtype(uint16).itemsize
                offset = z * length

                dt = numpy.dtype(uint16)
                dt = dt.newbyteorder('L')

                array = numpy.fromfile(file_name, offset=offset, count=length, dtype=dt)
            elif file_name.endswith('.blc'):
                raise NotImplementedError("This type of access is not yet supported")

            array = array.reshape(shape[1:])
            return array

        except FileNotFoundError:
            aprint(f"Could  not find file: {file_name} for array of shape: {shape} at z={z}")
            return numpy.zeros(shape[1:], dtype=uint16)

    def close(self):
        # Nothing to do...
        pass

    def channels(self):
        return list(self._channels)

    def shape(self, channel: str, time_point: int = 0) -> Sequence[int]:
        if (channel, time_point) in self._shapes:
            return (self._nb_time_points[channel],) + self._shapes[(channel, time_point)]
        else:
            return ()

    def dtype(self, channel: str):
        return numpy.uint16

    def info(self, channel: str = None) -> str:
        if channel:
            info_str = f"Channel: '{channel}', nb time points: {self.shape(channel)[0]}, shape: {self.shape(channel)[1:]} "
            return info_str
        else:
            return self.tree()

    def get_metadata(self):
        # TODO: implement this!
        return {}

    def append_metadata(self, metadata: dict):
        raise NotImplementedError("Method append_metadata is not available for a joined dataset!")

    def get_array(self, channel: str, per_z_slice: bool = True, wrap_with_dask: bool = False):

        # Lazy and memorized version of get_stack:
        lazy_get_stack = delayed(self.get_stack, pure=True)

        # Lazily load each stack for each time point:
        lazy_stacks = [lazy_get_stack(channel, time_point, per_z_slice) for time_point in self._time_points[channel]]

        # Construct a small Dask array for every lazy value:
        arrays = [array.from_delayed(lazy_stack,
                                     dtype=uint16,
                                     shape=self._channel_shape[channel])
                  for lazy_stack in lazy_stacks]

        stacked_array = array.stack(arrays, axis=0)  # Stack all small Dask arrays into one

        return stacked_array

    def get_stack(self, channel, time_point, per_z_slice=True, wrap_with_dask: bool = False):

        file_name = self._get_stack_file_name(channel, time_point)
        shape = self._shapes[(channel, time_point)]

        if per_z_slice and not self._is_compressed:

            lazy_get_slice_array_for_stack_file_and_z = delayed(self.cache.memoize(self._get_slice_array_for_stack_file_and_z), pure=True)

            # Lazily load each stack for each time point:
            lazy_stacks = [lazy_get_slice_array_for_stack_file_and_z(file_name, shape, z) for z in range(0, shape[0])]

            arrays = [array.from_delayed(lazy_stack,
                                         dtype=uint16,
                                         shape=shape[1:])
                      for lazy_stack in lazy_stacks]

            stack = array.stack(arrays, axis=0)

        else:
            stack = self._get_array_for_stack_file(file_name, shape=shape, dtype=uint16)

        return stack

    def add_channel(self, name: str, shape: Tuple[int, ...], dtype, enable_projections: bool = True, **kwargs) -> Any:
        raise NotImplementedError('Not implemented!')

    def get_projection_array(self, channel: str, axis: int, wrap_with_dask: bool = True) -> Any:
        return None

    def write_array(self, channel: str, array: numpy.ndarray):
        raise NotImplementedError('Not implemented!')

    def write_stack(self, channel: str, time_point: int, stack: numpy.ndarray):
        raise NotImplementedError('Not implemented!')

    def check_integrity(self, channels: Sequence[str]) -> bool:
        # TODO: actually implement!
        return True

    def _find_out_if_compressed(self, path):
        for root, dirs, files in os.walk(path):
            for file in files:
                if file.endswith(".blc"):
                    return True

        return False



from typing import Sequence, Optional

import numpy
from arbol.arbol import aprint, asection

from dexp.datasets.base_dataset import BaseDataset
from dexp.utils.slicing import slice_from_shape
from dexp.utils.misc import compute_num_workers

from joblib import Parallel, delayed


def dataset_copy(dataset: BaseDataset,
                 dest_path: str,
                 channels: Sequence[str],
                 slicing,
                 store: str = 'dir',
                 chunks: Optional[Sequence[int]] = None,
                 compression: str = 'zstd',
                 compression_level: int = 3,
                 overwrite: bool = False,
                 zerolevel: int = 0,
                 workers: int = 1,
                 workersbackend: Optional[str] = None,
                 check: bool = True,
                 stop_at_exception: bool = True,
                 ):

    # Create destination dataset:
    from dexp.datasets.zarr_dataset import ZDataset
    mode = 'w' + ('' if overwrite else '-')
    dest_dataset = ZDataset(dest_path, mode, store, parent=dataset)

    # Process each channel:
    for channel in dataset._selected_channels(channels):
        with asection(f"Copying channel {channel}:"):
            array = dataset.get_array(channel)

            aprint(f"Slicing with: {slicing}")
            out_shape, volume_slicing, time_points = slice_from_shape(array.shape, slicing)

            dtype = array.dtype
            dest_dataset.add_channel(name=channel,
                                     shape=out_shape,
                                     dtype=dtype,
                                     chunks=chunks,
                                     codec=compression,
                                     clevel=compression_level)

            def process(i):
                tp = time_points[i]
                try:
                    aprint(f"Processing time point: {i} ...")
                    tp_array = array[tp][volume_slicing]
                    if zerolevel != 0:
                        tp_array = numpy.array(tp_array)
                        tp_array = numpy.clip(tp_array, a_min=zerolevel, a_max=None, out=tp_array)
                        tp_array -= zerolevel
                    dest_dataset.write_stack(channel=channel,
                                             time_point=i,
                                             stack_array=tp_array)
                except Exception as error:
                    aprint(error)
                    aprint(f"Error occurred while copying time point {i} !")
                    import traceback
                    traceback.print_exc()
                    if stop_at_exception:
                        raise error

            if workers == 1:
                for i in range(len(time_points)):
                    process(i)
            else:
                n_jobs = compute_num_workers(workers, len(time_points))

                parallel = Parallel(n_jobs=n_jobs, backend=workersbackend)
                parallel(delayed(process)(i) for i in range(len(time_points)))

    # Dataset info:
    aprint(dest_dataset.info())

    # Check dataset integrity:
    if check:
        dest_dataset.check_integrity()

    # close destination dataset:
    dest_dataset.close()



from dask.distributed import Client
from typing import Tuple, List
import subprocess


def dask_init_client(port: int = 8786, n_gpus: int = 1, n_cpu_process: int = 1) \
        -> Tuple[Client, List[subprocess.Popen]]:
    commands = [
        f'dask-scheduler --port {port}',
        f'dask-worker tcp://localhost:{port} --nprocs {n_cpu_process} --nthreads 1 --resources CPU=1 --memory-limit=30e9 --name read-worker',
        f'dask-cuda-worker tcp://localhost:{port} --resources GPU={n_gpus} --name gpu-worker',
        f'dask-worker tcp://localhost:{port} --nprocs {n_cpu_process} --nthreads 1 --resources CPU=1 --memory-limit=30e9 --name write-worker',
    ]
    processes = [subprocess.Popen(c, shell=True) for c in commands]
    client = Client(f'127.0.0.1:{port}')
    return client, processes



from dexp.datasets.operations.crop import dataset_crop
import click
from arbol.arbol import aprint, asection

from dexp.cli.defaults import _default_clevel, _default_codec, _default_store
from dexp.cli.parsing import _parse_channels, _get_output_path, _parse_chunks
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.crop import dataset_crop


@click.command()
@click.argument('input_paths', nargs=-1)  # ,  help='input path'
@click.option('--output_path', '-o')  # , help='output path'
@click.option('--channels', '-c', default=None, help='List of channels, all channels when ommited.')
@click.option('--quantile', '-q', default=0.99, type=float, help='Quantile parameter for lower bound of brightness for thresholding.', show_default=True)
@click.option('--reference-channel', '-rc', default=None, help='Reference channel to estimate cropping. If no provided it picks the first one.')
@click.option('--store', '-st', default=_default_store, help='Zarr store: ‘dir’, ‘ndir’, or ‘zip’', show_default=True)
@click.option('--chunks', '-chk', default=None, help='Dataset chunks dimensions, e.g. (1, 126, 512, 512).')
@click.option('--codec', '-z', default=_default_codec, help='Compression codec: zstd for ’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ', show_default=True)
@click.option('--clevel', '-l', type=int, default=_default_clevel, help='Compression level', show_default=True)
@click.option('--overwrite', '-w', is_flag=True, help='Forces overwrite of target', show_default=True)
@click.option('--workers', '-wk', default=-4, help='Number of worker threads to spawn. Negative numbers n correspond to: number_of _cores / |n| ', show_default=True)  #
@click.option('--check', '-ck', default=True, help='Checking integrity of written file.', show_default=True)  #
def crop(input_paths,
         output_path,
         channels,
         quantile,
         reference_channel,
         store,
         chunks,
         codec,
         clevel,
         overwrite,
         workers,
         check):

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, '_crop')
    channels = _parse_channels(input_dataset, channels)
    if reference_channel is None:
        reference_channel = input_dataset.channels()[0]
    chunks = _parse_chunks(chunks)

    with asection(f"Cropping from: {input_paths} to {output_path} for channels: {channels}, "
                  f"using channel {reference_channel} as a reference."):

        dataset_crop(input_dataset,
                     output_path,
                     channels=channels,
                     reference_channel=reference_channel,
                     quantile=quantile,
                     store=store,
                     chunks=chunks,
                     compression=codec,
                     compression_level=clevel,
                     overwrite=overwrite,
                     workers=workers,
                     check=check)

        input_dataset.close()
        aprint("Done!")



import click
from arbol.arbol import aprint, asection

from dexp.cli.parsing import _parse_channels, _parse_slicing
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.extract_psf import dataset_extract_psf


@click.command(name='extract-psf')
@click.argument('input_paths', nargs=-1, required=True)
@click.option('--out-prefix-path', '-o', default='psf_', help='Output PSF file prefix')
@click.option('--channels', '-c', default=None,
             help='list of channels to extract the PSF, each channel will have their separater PSF (i.e., psf_c0.npy, psf_c1.npy')
@click.option('--slicing', '-s', default=None, help='dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ')
@click.option('--peak-threshold', '-pt', type=int, default=500, show_default=True,
              help='Peak valeu threshold for object (PSF) detection. Lower values are less consertaive and will detect more objects.')
@click.option('--similarity-threshold', '-st', type=float, default=0.5, show_default=True,
              help='Threshold of PSF selection given the similarity (cosine distance) to median PSF.')
@click.option('--psf_size', '-ps', type=int, default=35, show_default=True, help='Size (shape) of the PSF')
@click.option('--device', '-d', type=int, default=0, help='Sets the CUDA device id, e.g. 0,1,2. It works for only a single device!', show_default=True)
@click.option('--verbose', '-v', type=bool, is_flag=True, default=False, help='Flag to display intermediated results.')
def extract_psf(input_paths,
                out_prefix_path,
                channels,
                slicing,
                peak_threshold,
                similarity_threshold,
                psf_size,
                device,
                verbose,
                ):
    """Extracts the PSF from beads.
    """
    input_dataset, input_paths = glob_datasets(input_paths)
    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)

    with asection(f"Extracting PSF of dataset: {input_paths}, saving it with prefix: {out_prefix_path}, for channels: {channels}, slicing: {slicing} "):
        aprint(f"Device used: {device}")
        dataset_extract_psf(dataset=input_dataset,
                            dest_path=out_prefix_path,
                            channels=channels,
                            slicing=slicing,
                            peak_threshold=peak_threshold,
                            similarity_threshold=similarity_threshold,
                            psf_size=psf_size,
                            verbose=verbose,
                            device=device)



import gc
import math
import os
import threading
from typing import Any

import numpy
from arbol import aprint
from dask.array import Array

from dexp.processing.backends.backend import Backend
from dexp.utils import xpArray


class CupyBackend(Backend):
    """
    CupyBackend
    """
    _dexp_cuda_cluster = None
    _dexp_dask_client = None

    @staticmethod
    def num_devices():
        import GPUtil
        return len(GPUtil.getGPUs())

    @staticmethod
    def available_devices(order='first', maxLoad=math.inf, maxMemory=math.inf):
        """

        Parameters
        ----------
        order :     first --> select the GPU with the lowest ID (DEFAULT)
                    last --> select the GPU with the highest ID
                    random --> select a random available GPU
                    load --> select the GPU with the lowest load
                    memory --> select the GPU with the most memory available
        maxLoad :
        maxMemory

        Returns
        -------

        """
        # Set CUDA_DEVICE_ORDER so the IDs assigned by CUDA match those from nvidia-smi
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        import GPUtil
        return GPUtil.getAvailable(order=order,
                                   limit=math.inf,
                                   maxLoad=maxLoad,
                                   maxMemory=maxMemory,
                                   includeNan=False,
                                   excludeID=[],
                                   excludeUUID=[])

    device_locks = tuple(threading.Lock() for _ in range(num_devices.__func__()))

    def __init__(self,
                 device_id=0,
                 exclusive: bool = False,
                 enable_streaming: bool = True,
                 enable_memory_pool: bool = True,
                 enable_memory_pool_clearing: bool = True,
                 enable_unified_memory: bool = True,
                 enable_cub: bool = True,
                 enable_cutensor: bool = True,
                 enable_fft_planning: bool = True,
                 enable_dask_cuda_cluster: bool = False,
                 enable_dask_cuda_nvlink: bool = False,
                 ):
        """
        Instantiates a Cupy-based Image Processing backend

        Parameters
        ----------
        device_id : CUDA device id to use for allocation and compute
        exclusive : If True the access to this device is exclusive, no other backend context can access it (when using the context manager idiom)
        enable_memory_pool : Enables cupy memory pool. By default disabled, when enabled cupy tends to return out-of-memory exceptions when handling large arrays.
        enable_memory_pool_clearing : Enables the clearing of the memory pool upon calling 'clear_memory_pool'
        enable_cub : enables CUB accelerator
        enable_cutensor : enables cuTensor accelerator
        enable_fft_planning : enables FFT planning
        enable_dask_cuda_cluster : enables dask cuda cluster
        enable_dask_cuda_nvlink : enables nvlink
        """

        super().__init__()
        self.device_id = device_id
        self.exclusive = exclusive
        self.enable_memory_pool = enable_memory_pool
        self.enable_memory_pool_clearing = enable_memory_pool_clearing
        self.enable_streaming = enable_streaming
        self.enable_unified_memory = enable_unified_memory
        self.enable_cub = enable_cub
        self.enable_cutensor = enable_cutensor
        self.enable_fft_planning = enable_fft_planning
        self.enable_dask_cuda_cluster = enable_dask_cuda_cluster
        self.enable_dask_cuda_nvlink = enable_dask_cuda_nvlink

        self.mempool = None
        self.stream = None
        self._previous_allocator = None

        import cupy
        self.cupy_device: cupy.cuda.Device = cupy.cuda.Device(self.device_id)

        from cupy.cuda import cub, cutensor
        cub.available = enable_cub
        cutensor.available = enable_cutensor

        try:
            import cupy.cudnn
            aprint("CUDNN available!")
        except Exception as e:
            pass

        if not enable_fft_planning:
            import cupy
            cupy.fft.config.enable_nd_planning = False

        if enable_dask_cuda_cluster and CupyBackend._dexp_cuda_cluster is None:
            from dask_cuda import LocalCUDACluster
            from distributed import Client

            CupyBackend._dexp_cuda_cluster = LocalCUDACluster(enable_nvlink=enable_dask_cuda_nvlink)
            CupyBackend._dexp_dask_client = Client(CupyBackend._dexp_cuda_cluster)

        ## Important: Leave this, this is to make sure that some package works properly!
        exec("import cupyx.scipy.ndimage")
        exec("import cupyx.scipy.special")
        exec("import scipy.special")

    def copy(self, exclusive: bool = None):
        return CupyBackend(device_id=self.device_id,
                           exclusive=self.exclusive if exclusive is None else exclusive,
                           enable_streaming=self.enable_streaming,
                           enable_memory_pool=self.enable_memory_pool,
                           enable_memory_pool_clearing=self.enable_memory_pool_clearing,
                           enable_unified_memory=self.enable_unified_memory,
                           enable_cub=self.enable_cub,
                           enable_cutensor=self.enable_cutensor,
                           enable_fft_planning=self.enable_fft_planning,
                           enable_dask_cuda_cluster=self.enable_dask_cuda_cluster,
                           enable_dask_cuda_nvlink=self.enable_dask_cuda_nvlink)

    def __str__(self):
        free_mem = self.cupy_device.mem_info[0]
        total_mem = self.cupy_device.mem_info[0]
        percent = (100 * free_mem) // total_mem
        return (f"Cupy backend [device id:{self.device_id} "
                f"with {free_mem // (1024 * 1024)} MB ({percent}%) free memory out of {free_mem // (1024 * 1024)} MB, "
                f"compute:{self.cupy_device.compute_capability}, pci-bus-id:'{self.cupy_device.pci_bus_id}']")

    def __enter__(self):

        # lock device:
        if self.exclusive:
            CupyBackend.device_locks[self.device_id].acquire(blocking=True)

        # setup device:
        self.cupy_device.__enter__()

        # setup allocation:
        if self.enable_memory_pool:
            if self.mempool is None:
                import cupy
                from cupy.cuda.memory import SingleDeviceMemoryPool
                self.mempool = SingleDeviceMemoryPool(cupy.cuda.memory.malloc_managed if self.enable_unified_memory else None)
                from cupy.cuda import memory
            self._previous_allocator = memory._get_thread_local_allocator()
            memory._set_thread_local_allocator(self.mempool.malloc)
        else:
            from cupy.cuda import memory
            memory._set_thread_local_allocator(None)

        # setup stream:
        if self.enable_streaming:
            import cupy
            self.stream = cupy.cuda.stream.Stream(non_blocking=True)
            self.stream.__enter__()

        return super().__enter__()

    def __exit__(self, type, value, traceback):

        super().__exit__(type, value, traceback)

        # unset stream:
        if self.enable_streaming and self.stream is not None:
            self.stream.synchronize()
            self.stream.__exit__()

        # unset allocation:
        self.clear_memory_pool()
        if self._previous_allocator is not None:
            from cupy.cuda import memory
            memory._set_thread_local_allocator(self._previous_allocator)

        # unset device:
        self.cupy_device.__exit__()

        # unlock device:
        if self.exclusive:
            CupyBackend.device_locks[self.device_id].release()

    def synchronise(self):
        if self.stream is not None:
            self.stream.synchronize()

    def clear_memory_pool(self):

        if self.enable_memory_pool_clearing:
            if self.mempool is not None:
                n_free_blocks_before = self.mempool.n_free_blocks()
                used_before = self.mempool.used_bytes() // 1e9
                gc.collect()
                self.mempool.free_all_blocks(self.stream)
                n_free_blocks_after = self.mempool.n_free_blocks()
                used_after = self.mempool.used_bytes() // 1e9
                total_after = self.mempool.total_bytes() // 1e9

                aprint(f"Number of free blocks before and after release: {n_free_blocks_before}->{n_free_blocks_after}, used:{used_before}GB->{used_after}GB, total:{total_after}GB ")
            else:
                aprint("Warning: default cupy memory pool is 'None'")
        else:
            aprint("Memory pool clearing not enabled!")

        super().clear_memory_pool()

    def _to_numpy(self, array: xpArray, dtype=None, force_copy: bool = False) -> numpy.ndarray:
        import cupy

        if isinstance(array, Array):
            return self._to_numpy(array.compute(), dtype=dtype, force_copy=force_copy)

        elif cupy.get_array_module(array) == cupy:
            array = cupy.asnumpy(array)

        if dtype:
            return numpy.asarray(array).astype(dtype, copy=force_copy)
        elif force_copy:
            return numpy.asarray(array.copy())
        else:
            return numpy.asarray(array)

    def _to_backend(self, array: xpArray, dtype=None, force_copy: bool = False) -> Any:

        import cupy

        if isinstance(array, Array):
            return self._to_backend(array.compute(), dtype=dtype, force_copy=force_copy)

        elif cupy.get_array_module(array) == cupy:
            if dtype:
                return array.astype(dtype, copy=force_copy)
            elif force_copy:
                return array.copy()
            else:
                return array
        else:
            array = self.to_numpy(array)
            with self.cupy_device:
                return cupy.asarray(array, dtype=dtype)

    def _get_xp_module(self, array: xpArray = None) -> Any:
        if array is None:
            import cupy
            return cupy
        else:
            import cupy
            return cupy.get_array_module(array)

    def _get_sp_module(self, array: xpArray = None) -> Any:
        if array is None:
            import cupyx
            return cupyx.scipy
        else:
            import cupyx
            return cupyx.scipy.get_array_module(array)



import os
from typing import Sequence, Optional

import numpy
from arbol.arbol import aprint
from arbol.arbol import asection
from joblib import Parallel, delayed

from dexp.datasets.base_dataset import BaseDataset
from dexp.processing.backends.best_backend import BestBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.registration.model.sequence_registration_model import SequenceRegistrationModel
from dexp.processing.registration.sequence import image_stabilisation
from dexp.processing.registration.sequence_proj import image_stabilisation_proj_
from dexp.utils.misc import compute_num_workers


def _compute_model(
        input_dataset: BaseDataset,
        channel: str,
        slicing,
        max_range: int,
        min_confidence: float,
        enable_com: bool,
        quantile: float,
        tolerance: float,
        order_error: float,
        order_reg: float,
        alpha_reg: float,
        phase_correlogram_sigma: float,
        denoise_input_sigma: float,
        log_compression: bool,
        edge_filter: bool,
        detrend: bool = False,
        maxproj: bool = True,
        device: int = 0,
        workers: int = 1,
        debug_output=None,
    ) -> SequenceRegistrationModel:

    # get channel array:
    array = input_dataset.get_array(channel, per_z_slice=False, wrap_with_dask=True)

    # Perform slicing:
    if slicing is not None:
        array = array[slicing]

    # Shape and chunks for array:
    ndim = array.ndim

    workers = compute_num_workers(workers, array.shape[0])

    if not maxproj:
        with BestBackend(device, enable_unified_memory=True):
            model = image_stabilisation(
                image=array,
                axis=0,
                detrend=detrend,
                preload_images=False,
                max_range=max_range,
                min_confidence=min_confidence,
                enable_com=enable_com,
                quantile=quantile,
                tolerance=tolerance,
                order_error=order_error,
                order_reg=order_reg,
                alpha_reg=alpha_reg,
                sigma=phase_correlogram_sigma,
                denoise_input_sigma=denoise_input_sigma,
                log_compression=log_compression,
                edge_filter=edge_filter,
                internal_dtype=numpy.float16,
                workers=workers,
                debug_output=debug_output,
            )
            model.to_numpy()

    else:  # stabilized with maximum intensity projection
        projections = []
        for axis in range(array.ndim - 1):
            projection = input_dataset.get_projection_array(
                channel=channel, axis=axis, wrap_with_dask=False
            )
            if slicing is not None:
                projection = projection[slicing]

            proj_axis = list(1 + a for a in range(array.ndim - 1) if a != axis)
            projections.append((*proj_axis, projection))

        # Perform stabilisation:
        with BestBackend(device, enable_unified_memory=True):
            model = image_stabilisation_proj_(
                projections=projections,
                max_range=max_range,
                min_confidence=min_confidence,
                enable_com=enable_com,
                quantile=quantile,
                tolerance=tolerance,
                order_error=order_error,
                order_reg=order_reg,
                alpha_reg=alpha_reg,
                sigma=phase_correlogram_sigma,
                denoise_input_sigma=denoise_input_sigma,
                log_compression=log_compression,
                edge_filter=edge_filter,
                ndim=ndim - 1,
                internal_dtype=numpy.float16,
                debug_output=debug_output,
                detrend=detrend,
            )
            model.to_numpy()

    return model


def dataset_stabilize(dataset: BaseDataset,
                      output_path: str,
                      channels: Sequence[str],
                      reference_channel: Optional[str] = None,
                      slicing=None,
                      zarr_store: str = "dir",
                      compression_codec: str = "zstd",
                      compression_level: int = 3,
                      overwrite: bool = False,
                      max_range: int = 7,
                      min_confidence: float = 0.5,
                      enable_com: bool = False,
                      quantile: float = 0.5,
                      tolerance: float = 1e-7,
                      order_error: float = 2.0,
                      order_reg: float = 1.0,
                      alpha_reg: float = 0.1,
                      phase_correlogram_sigma: float = 2,
                      denoise_input_sigma: float = 1.5,
                      log_compression: bool = True,
                      edge_filter: bool = False,
                      pad: bool = True,
                      integral: bool = True,
                      detrend: bool = False,
                      maxproj: bool = True,
                      workers: int = -1,
                      workers_backend: str = "threading",
                      device: int = 0,
                      check: bool = True,
                      stop_at_exception: bool = True,
                      debug_output=None):
    """
    Takes an input dataset and performs image stabilisation and outputs a stabilised dataset with given selected slice & channels in Zarr format with given store type, compression, etc...

    Parameters
    ----------
    dataset: Input dataset
    output_path: Output path for Zarr storage
    channels: selected channels
    reference_channel: use this channel to compute the stabilization model and apply to every channel.
    slicing: selected array slicing
    zarr_store: type of store, can be 'dir', 'ndir', or 'zip'
    compression_codec: compression codec to be used ('zstd', 'blosclz', 'lz4', 'lz4hc', 'zlib' or 'snappy').
    compression_level: An integer between 0 and 9 specifying the compression level.
    overwrite: overwrite output dataset if already exists
    max_range: maximal distance, in time points, between pairs of images to registrate.
    min_confidence: minimal confidence below which pairwise registrations are rejected for the stabilisation.
    enable_com: enable center of mass fallback when standard registration fails.
    quantile: quantile to cut-off background in center-of-mass calculation
    tolerance: tolerance for linear solver.
    order_error: order for linear solver error term.
    order_reg: order for linear solver regularisation term.
    alpha_reg: multiplicative coefficient for regularisation term.
    phase_correlogram_sigma: sigma for Gaussian smoothing of phase correlogram, zero to disable.
    denoise_input_sigma : Uses a Gaussian filter to denoise input images, zero to disable.
    log_compression : Applies the function log1p to the images to compress high-intensities (usefull when very (too) bright structures are present in the images, such as beads)
    edge_filter : apply sobel edge filter to input images.
    pad: pad input dataset.
    integral: Set to True to only allow integral translations, False to allow subpixel accurate translations (induces blur!).
    detrend: removes linear detrend from stabilized image.
    maxproj: uses maximum intensity projection to compute the volume stabilization.
    workers: number of workers, if -1 then the number of workers == number of cores or devices
    workers_backend: What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread)
    device: Sets the CUDA devices id, e.g. 0,1,2
    check: Checking integrity of written file.
    stop_at_exception: True to stop as soon as there is an exception during processing.
    """

    from dexp.datasets.zarr_dataset import ZDataset
    mode = 'w' + ('' if overwrite else '-')
    dest_dataset = ZDataset(output_path, mode, zarr_store, parent=dataset)

    model = None
    if reference_channel is not None:
        if reference_channel not in dataset.channels():
            raise ValueError(f'Reference channel {reference_channel} not found.')
        model = _compute_model(
            input_dataset=dataset,
            channel=reference_channel,
            slicing=slicing,
            max_range=max_range,
            min_confidence=min_confidence,
            enable_com=enable_com,
            quantile=quantile,
            tolerance=tolerance,
            order_error=order_error,
            order_reg=order_reg,
            alpha_reg=alpha_reg,
            phase_correlogram_sigma=phase_correlogram_sigma,
            denoise_input_sigma=denoise_input_sigma,
            log_compression=log_compression,
            edge_filter=edge_filter,
            detrend=detrend,
            maxproj=maxproj,
            device=device,
            workers=workers,
            debug_output=debug_output,
        )

    for channel in dataset._selected_channels(channels):
        if reference_channel is None:
            channel_model = _compute_model(
                input_dataset=dataset,
                channel=channel,
                slicing=slicing,
                max_range=max_range,
                min_confidence=min_confidence,
                enable_com=enable_com,
                quantile=quantile,
                tolerance=tolerance,
                order_error=order_error,
                order_reg=order_reg,
                alpha_reg=alpha_reg,
                phase_correlogram_sigma=phase_correlogram_sigma,
                denoise_input_sigma=denoise_input_sigma,
                log_compression=log_compression,
                edge_filter=edge_filter,
                detrend=detrend,
                maxproj=maxproj,
                device=device,
                workers=workers,
                debug_output=debug_output,
            )
        else:
            channel_model = model

        metadata = dataset.get_metadata()
        if 'dt' in metadata.get(channel, {}):
            prev_displacement = channel_model.total_displacement()
            channel_model = channel_model.reduce(int(round(metadata[channel]['dt'])))
            displacement = channel_model.total_displacement()

            assert len(displacement) == 3
            axes = ('tz', 'ty', 'tx')
            asection('Translation displacement after reduction')
            for name, prev, current in zip(axes, prev_displacement, displacement):
                # each displacement is a tuple (min, max) of displacement
                # minimum displacement different results in translations shift from reference channel
                metadata[channel][name] = current[0] - prev[0]
                aprint(f'{name.upper()}: {current[0] - prev[0]}')
            dest_dataset.append_metadata(metadata)

        array = dataset.get_array(channel, per_z_slice=False, wrap_with_dask=True)
        if slicing is not None:
            array = array[slicing]

        shape = array.shape
        dtype = array.dtype
        nb_timepoints = shape[0]

        if len(channel_model) != nb_timepoints:
            aprint(f'WARNING: Number of time points ({nb_timepoints}) and registration models ({len(channel_model)})'
                   f'does not match. Using the smallest one.')
            if nb_timepoints > len(channel_model):
                nb_timepoints = len(channel_model)
            # else: it is not necessary to change the model sequence length

        # Shape of the resulting array:
        padded_shape = (nb_timepoints,) + channel_model.padded_shape(shape[1:])

        dest_dataset.add_channel(name=channel,
                                 shape=padded_shape,
                                 dtype=dtype,
                                 codec=compression_codec,
                                 clevel=compression_level)

        # definition of function that processes each time point:
        def process(tp):
            try:
                with asection(f"Processing time point: {tp}/{nb_timepoints} ."):
                    with asection(f"Loading stack"):
                        tp_array = array[tp].compute()

                    with NumpyBackend():
                        with asection(f"Applying model..."):
                            tp_array = channel_model.apply(tp_array, index=tp, pad=pad, integral=integral)

                    with asection(f"Saving stabilized stack for time point {tp}/{nb_timepoints}, shape:{tp_array.shape}, dtype:{array.dtype}"):
                        dest_dataset.write_stack(channel=channel,
                                                 time_point=tp,
                                                 stack_array=tp_array)

            except Exception as error:
                aprint(error)
                aprint(f"Error occurred while processing time point {tp}/{nb_timepoints} !")
                import traceback
                traceback.print_exc()
                if stop_at_exception:
                    raise error

        # start jobs:
        if workers == 1:
            for tp in range(0, nb_timepoints):
                process(tp)
        else:
            n_jobs = compute_num_workers(workers, nb_timepoints)
            Parallel(n_jobs=n_jobs, backend=workers_backend)(
                delayed(process)(tp) for tp in range(0, nb_timepoints)
            )

    # printout output dataset info:
    aprint(dest_dataset.info())
    if check:
        dest_dataset.check_integrity()

    # Dataset info:
    aprint(dest_dataset.info())

    # Check dataset integrity:
    if check:
        dest_dataset.check_integrity()

    # close destination dataset:
    dest_dataset.close()



from typing import List

import numpy as np
from numpy.typing import ArrayLike

from itertools import combinations

from dexp.utils import xpArray
from dexp.processing.backends.backend import Backend


__all__ = [
    'first_derivative_func',
    'first_derivative_kernels',
    'second_derivative_func',
    'second_derivative_kernels',
    'derivative_axes',
]


def line_derivative_kernels(dim: int, template: ArrayLike) -> List[xpArray]:
    kernels = []
    for axis in range(dim):
        shape = np.ones(dim, dtype=int)
        shape[axis] = 3
        D = np.zeros(shape, dtype=np.float32)
        slicing = tuple(slice(None) if i == axis else 0 for i in range(dim))
        D[slicing] = template
        kernels.append(D)
    return kernels


def diagonal_derivative_kernels(dim: int, template: ArrayLike) -> List[xpArray]:
    kernels = []
    for axes in combinations(range(dim), 2):
        shape = np.ones(dim, dtype=int)
        shape[list(axes)] = 3
        D = np.zeros(shape, dtype=np.float32)
        slicing = tuple(slice(None) if i in axes else 0 for i in range(dim))
        D[slicing] = template
        kernels.append(D)

    return kernels


def first_derivative_kernels(dim: int) -> List[xpArray]:
    kernels = []
    line = np.array([1, -1, 0])
    kernels += line_derivative_kernels(dim, line)
    diagonal = np.array([[1,  0, 0],
                         [0, -1, 0],
                         [0,  0, 0]])
    kernels += diagonal_derivative_kernels(dim, diagonal)
    return kernels


def second_derivative_kernels(dim: int) -> List[xpArray]:
    kernels = []
    line = np.array([1, -2, 1])
    kernels += line_derivative_kernels(dim, line)
    diagonal = np.array([[1,  0, 0],
                         [0, -2, 0],
                         [0,  0, 1]])
    kernels += diagonal_derivative_kernels(dim, diagonal)
    return kernels


def derivative_axes(dim: int) -> List[int]:
    # Must be kept in the same order as the other derivative functions
    return list(range(dim)) + list(combinations(range(dim), 2))


def first_derivative_func(array: xpArray, axes: int, transpose: bool) -> xpArray:
    xp = Backend.get_xp_module()

    left_slice = [slice(None) for _ in range(array.ndim)]
    right_slice = [slice(None) for _ in range(array.ndim)]

    if isinstance(axes, int):
        axes = (axes,)

    for axis in axes:
        left_slice[axis] = slice(None, -1)
        right_slice[axis] = slice(1, None)

    if transpose:
        left_slice, right_slice = right_slice, left_slice
    
    left_slice = tuple(left_slice)
    right_slice = tuple(right_slice)

    out = xp.zeros_like(array)
    out[right_slice] = array[left_slice] - array[right_slice]

    return out


def second_derivative_func(array: xpArray, axes: int, transpose: bool) -> xpArray:
    return -1 * first_derivative_func(
        first_derivative_func(array, axes, transpose), axes, not transpose
    )



import pytest
import numpy as np
from numpy.testing import assert_allclose

from dexp.processing.deconvolution.admm_utils import *
import scipy.ndimage as ndi


@pytest.mark.parametrize(
    'derivative_kernels, derivative_func', [
        (first_derivative_kernels, first_derivative_func),
        (second_derivative_kernels, second_derivative_func),
    ]
)
def test_derivatives(derivative_kernels, derivative_func):
    image = np.random.randint(256, size=(64, 64, 64))

    # ignoring the border, the line kernel is exact
    # the diagonal kernel ignores the borders
    centering = tuple((slice(1, 63) for _ in range(image.ndim)))

    Ks = derivative_kernels(image.ndim)
    Daxes = derivative_axes(image.ndim)

    for K, axes in zip(Ks, Daxes):
        convolved = ndi.convolve(image, K, mode='nearest')[centering]
        finite_dif = derivative_func(image, axes, transpose=True)[centering]

        assert_allclose(convolved, finite_dif, err_msg=f'Axes {axes}', verbose=True)

        correlated = ndi.correlate(image, K, mode='nearest')[centering]
        finite_dif = derivative_func(image, axes, transpose=False)[centering]

        assert_allclose(correlated, finite_dif, err_msg=f'Axes {axes}', verbose=True)



from arbol import aprint
import multiprocessing
from numcodecs import blosc


def config_blosc():
    _cpu_count = multiprocessing.cpu_count() // 2
    _nb_threads = max(1, _cpu_count)
    blosc.use_threads = True
    blosc.set_nthreads(_nb_threads)
    aprint(f"Configured the number of threads used by BLOSC: {blosc.get_nthreads()}")



from typing import Sequence, Optional, Tuple

import zarr
import numpy as np
from numpy.typing import ArrayLike
from arbol.arbol import aprint, asection

from dexp.datasets.base_dataset import BaseDataset
from dexp.processing.backends.best_backend import BestBackend
from dexp.processing.backends.backend import Backend
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.utils.misc import compute_num_workers
from scipy import ndimage as ndi


from joblib import Parallel, delayed


def _estimate_crop(array: ArrayLike, quantile: float = 0.99) -> Sequence[Tuple[int]]:
    window_size = 31
    step = 4
    shape = array.shape
    with BestBackend():
        xp = Backend.get_xp_module()
        array = Backend.to_backend(array[::step,::step,::step], dtype=xp.float16)
        array = xp.clip(array - xp.mean(array), 0, None)  # removing background noise
        kernel = xp.ones((window_size, window_size, window_size)) / (window_size ** 3)
        kernel = kernel.astype(xp.float16)
        array = fft_convolve(array, kernel, in_place=True)
        lower = xp.quantile(array, quantile)
        aprint('Estimated lower threshold', lower)
        array = array > lower
        array, _ = ndi.label(Backend.to_numpy(array))
        slices = ndi.find_objects(array)
    
    largest_slice = None
    largest_size = 0
    for slicing in slices:
        size = np.prod(tuple((s.stop - s.start) for s in slicing))
        if size > largest_size:
            largest_slice = slicing
            largest_size = size

    if largest_slice is None:
        raise RuntimeError('Could not detect any objects')

    return tuple((s.start * step, min(s.stop * step, d)) # fixing possible mismatch due to step
        for s, d in zip(largest_slice, shape))


def compute_crop_slicing(array: zarr.Array, time_points: Sequence[int], quantile: float) -> Sequence[slice]:
    # N x D x 2
    ranges = np.array(tuple(_estimate_crop(array[t], quantile) for t in time_points))
    lower = np.min(ranges[..., 0], axis=0)
    upper = np.max(ranges[..., 1], axis=0)
    return tuple(slice(int(l), int(u)) for l, u in zip(lower, upper))


def dataset_crop(dataset: BaseDataset,
                 dest_path: str,
                 reference_channel: str,
                 channels: Sequence[str],
                 quantile: float,
                 store: str = 'dir',
                 chunks: Optional[Sequence[int]] = None,
                 compression: str = 'zstd',
                 compression_level: int = 3,
                 overwrite: bool = False,
                 workers: int = 1,
                 check: bool = True,
                 stop_at_exception: bool = True,
                 ):

    # Create destination dataset:
    from dexp.datasets.zarr_dataset import ZDataset
    mode = 'w' + ('' if overwrite else '-')
    dest_dataset = ZDataset(dest_path, mode, store, parent=dataset)

    with asection("Estimating region of interest"):
        nb_time_pts = dataset.nb_timepoints(reference_channel)
        slicing = compute_crop_slicing(dataset.get_array(reference_channel), [0, nb_time_pts // 2, nb_time_pts-1], quantile)
        aprint('Estimated slicing of', slicing)
        volume_shape = tuple(s.stop - s.start for s in slicing)
        translation = {
            k: s.start for k, s in zip(('tz', 'ty', 'tx'), slicing)
        }
        dest_dataset.append_metadata(translation)

    # Process each channel:
    for channel in dataset._selected_channels(channels):
        with asection(f"Cropping channel {channel}:"):
            array = dataset.get_array(channel)

            dtype = array.dtype
            dest_dataset.add_channel(name=channel,
                                     shape=(len(array),) + volume_shape,
                                     dtype=dtype,
                                     chunks=chunks,
                                     codec=compression,
                                     clevel=compression_level)

            def process(tp):
                try:
                    aprint(f"Processing time point: {tp} ...")
                    tp_array = array[tp][slicing]
                    dest_dataset.write_stack(channel=channel,
                                             time_point=tp,
                                             stack_array=tp_array)
                except Exception as error:
                    aprint(error)
                    aprint(f"Error occurred while copying time point {tp} !")
                    import traceback
                    traceback.print_exc()
                    if stop_at_exception:
                        raise error

            if workers == 1:
                for i in range(len(array)):
                    process(i)
            else:
                n_jobs = compute_num_workers(workers, len(array))

                parallel = Parallel(n_jobs=n_jobs)
                parallel(delayed(process)(i) for i in range(len(array)))

    # Dataset info:
    aprint(dest_dataset.info())

    # Check dataset integrity:
    if check:
        dest_dataset.check_integrity()

    # close destination dataset:
    dest_dataset.close()



import math

import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.utils import xpArray


def classic_deskew(image: xpArray,
                   depth_axis: int,
                   lateral_axis: int,
                   dx: float,
                   dz: float,
                   angle: float,
                   camera_orientation: int = 0,
                   flip_depth_axis: bool = False,
                   epsilon: float = 1e-2,
                   order: int = 1,
                   padding: bool = True,
                   copy: bool = True,
                   internal_dtype=None
                   ):
    """ Classic Deskewing.

    Parameters
    ----------
    image  : input image (skewed 3D stack)
    depth_axis  : Depth axis.
    lateral_axis  : Lateral axis.
    dz     : float, scanning step (stage or galvo scanning step, not the same as the distance between the slices)
    dx     : float, pixel size of the camera
    angle  : float, incident angle of the light sheet, angle between the light sheet and the optical axis in degrees
    camera_orientation : Camera orientation correction expressed as a number of 90 deg rotations to be performed per 2D image in stack -- if required.
    flip_depth_axis : Flips depth axis to deskew in the opposite orientation (True for view 0 and False for view 1)
    epsilon : relative tolerance to non-integral shifts: if the shift is not _exactly_ integer, to which extent do we actually tolerate that when switching between the 'roll' method and full interpolation approach.
    order : interpolation order (when necessary only!)
    padding : Apply padding, or not.
    copy : Set to True to force copy of images.
    internal_dtype : internal dtype to perform computation

    Returns
    -------
    Deskewed image

    """

    # computes the dimension-less parameters:
    shift = dx * math.sin(angle * math.pi / 180) / dz
    lateral_scaling = math.cos(angle * math.pi / 180)

    image = classic_deskew_dimensionless(image=image,
                                         shift=shift,
                                         depth_axis=depth_axis,
                                         lateral_axis=lateral_axis,
                                         flip_depth_axis=flip_depth_axis,
                                         lateral_scaling=lateral_scaling,
                                         camera_orientation=camera_orientation,
                                         epsilon=epsilon,
                                         order=order,
                                         padding=padding,
                                         copy=copy,
                                         internal_dtype=internal_dtype)

    return image


def classic_deskew_dimensionless(image: xpArray,
                                 depth_axis: int,
                                 lateral_axis: int,
                                 shift: float,
                                 lateral_scaling: float = 1.0,
                                 flip_depth_axis: bool = False,
                                 camera_orientation: int = 0,
                                 epsilon: float = 1e-2,
                                 order: int = 1,
                                 padding: bool = False,
                                 copy: bool = True,
                                 internal_dtype=None
                                 ):
    """ Classic Deskewing with the dimensionless parametrisation.

    Parameters
    ----------
    image  : input image (skewed 3D stack)
    depth_axis  : Depth axis.
    lateral_axis  : Lateral axis.
    shift  : Shift to apply along lateral axis per plane stacked along the depth axis.
    lateral_scaling : Scaling necessary along the lateral axis to preserve voxel dimensions.
    flip_depth_axis : Flips depth axis to deskew in the opposite orientation (True for view 0 and False for view 1)
    camera_orientation : Camera orientation correction expressed as a number of 90 deg rotations to be performed per 2D image in stack -- if required.
    epsilon : relative tolerance to non-integral shifts: if the shift is not _exactly_ integer, to which extent do we actually tolerate that when switching between the 'roll' method and full interpolation approach.
    order : interpolation order (when necessary only!)
    padding : Apply padding, or not.
    copy : Set to True to force copy of images.
    internal_dtype : internal dtype to perform computation

    Returns
    -------
    Deskewed image

    """

    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if internal_dtype is None:
        internal_dtype = xp.float16

    if type(Backend.current()) is NumpyBackend and internal_dtype == xp.float16:
        internal_dtype = xp.float32

    # Save original dtype:
    original_dtype = image.dtype

    # Move images to backend.
    image = Backend.to_backend(image, dtype=internal_dtype, force_copy=copy)

    # First we compute the permutation that will reorder the axis so that the depth and lateral axis are the first axis in the image:
    permutation = (depth_axis, lateral_axis) + tuple(
        axis
        for axis in range(image.ndim)
        if axis not in [depth_axis, lateral_axis]
    )
    permutation = numpy.asarray(permutation)
    inverse_permutation = numpy.argsort(permutation)

    # We apply the permutation:
    image = xp.transpose(image, axes=permutation)

    # Camera orientation correction, if required:
    image = xp.rot90(image, k=camera_orientation, axes=(1, 2))

    if flip_depth_axis:
        image = xp.flip(image, axis=0)

    # Image height and depth:
    depth = image.shape[0]
    height = image.shape[1]

    # Padding:
    if padding:
        pad_width_z = int(round(abs(shift) * height / 2))
        pad_width = ((pad_width_z, pad_width_z), (0, 0), (0, 0),)
        image = xp.pad(image, pad_width=pad_width)

    # Is the shift integral?
    integral_shift = abs(shift - int(shift)) / shift < epsilon

    if integral_shift:
        # Shift is integral:
        for hi in range(height):
            image[:, hi, ...] = xp.roll(image[:, hi, ...], shift=shift * hi, axis=0)
    else:
        # Shift is not integral, we go full interpolation
        # Deskew matrix:
        alpha = 1
        matrix = xp.asarray([[1, -shift, 0], [0, 1, 0], [0, 0, 1]])

        # deskew:
        image = sp.ndimage.affine_transform(image, matrix, order=order)

    # rescale the data along lateral axis to preserve voxel dimensions:
    # if lateral_scaling != 1.0:
    #     zoom = [1, ] * 3
    #     zoom[lateral_axis] = lateral_scaling
    #     image = sp.ndimage.zoom(image, zoom=zoom, order=order)

    # flip along axis z
    if flip_depth_axis:
        image = xp.flip(image, axis=0)

    # We apply the inverse permutation:
    image = xp.transpose(image, axes=inverse_permutation)

    # Cast back to original dtype, if needed:
    image = image.astype(original_dtype, copy=False)

    return image



from typing import List, Tuple, Optional, Dict
import warnings

import numpy
import pandas as pd
from arbol import asection, aprint

from dexp.processing.backends.backend import Backend
from dexp.processing.equalise.equalise_intensity import equalise_intensity
from dexp.processing.filters.butterworth_filter import butterworth_filter
from dexp.processing.fusion.dct_fusion import fuse_dct_nd
from dexp.processing.fusion.dft_fusion import fuse_dft_nd
from dexp.processing.fusion.tg_fusion import fuse_tg_nd
from dexp.processing.registration.model import RegistrationModel
from dexp.processing.registration.translation_nd import register_translation_nd
from dexp.processing.registration.translation_nd_proj import register_translation_proj_nd
from dexp.processing.restoration.clean_dark_regions import clean_dark_regions
from dexp.processing.restoration.dehazing import dehaze
from dexp.utils import xpArray
from dexp.processing.multiview_lightsheet.fusion.basefusion import BaseFusion


class SimViewFusion(BaseFusion):
    def __init__(self,
                 registration_model: Optional[RegistrationModel],
                 equalise: bool,
                 equalisation_ratios: List[Optional[float]],
                 zero_level: float,
                 clip_too_high: int,
                 fusion: str,
                 fusion_bias_exponent: int,
                 fusion_bias_strength_i: float,
                 fusion_bias_strength_d: float,
                 dehaze_before_fusion: bool,
                 dehaze_size: int,
                 dehaze_correct_max_level: bool,
                 dark_denoise_threshold: int,
                 dark_denoise_size: int,
                 butterworth_filter_cutoff: float,
                 flip_camera1: bool,
                 internal_dtype: numpy.dtype = numpy.float16,
                 pad: bool = True,
                 ):
        super().__init__(registration_model, equalise, equalisation_ratios, zero_level, clip_too_high, fusion,
                         dehaze_before_fusion, dehaze_size, dehaze_correct_max_level,
                         dark_denoise_threshold, dark_denoise_size, butterworth_filter_cutoff, internal_dtype)

        self._fusion_bias_exponent = fusion_bias_exponent
        self._fusion_bias_strength_i = fusion_bias_strength_i
        self._fusion_bias_strength_d = fusion_bias_strength_d
        self._flip_camera1 = flip_camera1
    
    def _preprocess_single_view(self, view: xpArray, camera: int, lightsheet: int, flip: bool) -> xpArray:
        xp = Backend.get_xp_module()

        with asection(f"Moving C{camera}L{lightsheet} and to backend storage and converting to {self._internal_dtype} ..."):
            view = Backend.to_backend(view, dtype=self._internal_dtype, force_copy=False)

        if self._clip_too_high > 0:
            with asection(f"Clipping intensities above {self._clip_too_high} for C{camera}L0"):
                xp.clip(view, a_min=0, a_max=self._clip_too_high, out=view)

        if self._dehaze_size > 0 and self._dehaze_before_fusion:
            with asection(f"Dehaze C{camera}L{lightsheet} ..."):
                view = dehaze(view,
                              size=self._dehaze_size,
                              minimal_zero_level=0,
                              correct_max_level=True)

        if flip:
            view = xp.flip(view, -1).copy()

        return view

    def _preprocess_and_fuse_illumination_views(self,
                                                view0: xpArray,
                                                view1: xpArray,
                                                flip: bool,
                                                camera: int) -> xpArray:
        
        view0 = self._preprocess_single_view(view0, camera, 0, flip)
        view1 = self._preprocess_single_view(view1, camera, 1, flip)

        if self._equalise:
            with asection(f"Equalise intensity of C{camera}L0 relative to C{camera}L1 ..."):
                view0, view1, ratio = equalise_intensity(view0, view1,
                                                         zero_level=self._zero_level,
                                                         correction_ratio=self._equalisation_ratios[camera],
                                                         copy=False)
                aprint(f"Equalisation ratio: {ratio}")
                self._equalisation_ratios[camera] = ratio

        with asection(f"Fuse illumination views C{camera}L0 and C{camera}L1..."):
            fused_view = self._fuse_illumination_views(view0, view1)

        return fused_view

    def preprocess(self, C0L0: xpArray, C0L1: Optional[xpArray] = None,
                   C1L0: Optional[xpArray] = None, C1L1: Optional[xpArray] = None) \
        -> Tuple[xpArray, Optional[xpArray]]:

        if C0L1 is not None:
            self._match_input(C0L0, C0L1)
        if C1L0 is not None:
            self._match_input(C0L0, C1L0)
        if C1L1 is not None:
            self._match_input(C0L0, C1L1)

        if C0L1 is None:
            # processing a single light sheet from a single camera
            C0Lx = self._preprocess_single_view(C0L0, camera=0, lightsheet=0, flip=False)
        else:
            C0Lx = self._preprocess_and_fuse_illumination_views(C0L0, C0L1, flip=False, camera=0)

        if C1L0 is None and C1L1 is None:
            # fusing only images from a single camera
            return C0Lx, None

        if C1L1 is None:
            # processing a single light sheet from a single camera
            C1Lx = self._preprocess_single_view(C1L0, camera=1, lightsheet=0, flip=self._flip_camera1)
        else:
            C1Lx = self._preprocess_and_fuse_illumination_views(C1L0, C1L1, flip=self._flip_camera1, camera=1)

        if self._equalise:
            with asection(f"Equalise intensity of C0Lx relative to C1Lx ..."):
                C0Lx, C1Lx, ratio = equalise_intensity(C0Lx, C1Lx,
                                                       zero_level=0,
                                                       correction_ratio=self._equalisation_ratios[2],
                                                       copy=False)
                aprint(f"Equalisation ratio: {ratio}")
                self._equalisation_ratios[2] = ratio

        return C0Lx, C1Lx

    def postprocess(self, CxLx: xpArray) -> xpArray:
        if self._dehaze_size > 0 and not self._dehaze_before_fusion:
            with asection(f"Dehaze CxLx ..."):
                CxLx = dehaze(CxLx,
                              size=self._dehaze_size,
                              minimal_zero_level=0,
                              correct_max_level=self._dehaze_correct_max_level)

        if self._dark_denoise_threshold > 0:
            with asection(f"Denoise dark regions of CxLx..."):
                CxLx = clean_dark_regions(CxLx,
                                          size=self._dark_denoise_size,
                                          threshold=self._dark_denoise_threshold)

        if 0 < self._butterworth_filter_cutoff < 1:
            with asection(f"Filter output using a Butterworth filter"):
                cutoffs = (self._butterworth_filter_cutoff,) * CxLx.ndim
                CxLx = butterworth_filter(CxLx, shape=(31, 31, 31), cutoffs=cutoffs, cutoffs_in_freq_units=False)

        return CxLx

    def fuse(self, C0Lx: xpArray, C1Lx: xpArray, pad: bool) -> xpArray:
        if self._registration_model is None:
            raise RuntimeError('Registration must be computed beforehand.')

        with asection(f"Register_stacks C0Lx and C1Lx ..."):
            C0Lx, C1Lx = self._registration_model.apply_pair(C0Lx, C1Lx, pad=pad)

        with asection(f"Fuse detection views C0lx and C1Lx..."):
            CxLx = self._fuse_detection_views(C0Lx, C1Lx)

        return CxLx

    def __call__(self, C0L0: xpArray, C0L1: Optional[xpArray] = None,
                 C1L0: Optional[xpArray] = None, C1L1: Optional[xpArray] = None,
                 pad: bool = False) -> xpArray:

        original_dtype = C0L0.dtype
        xp = Backend.current().get_xp_module()

        C0Lx, C1Lx = self.preprocess(C0L0, C0L1, C1L0, C1L1)
        if C1Lx is None:
            # fusing only two views from two different light sheets
            CxLx = C0Lx
        else:
            CxLx = self.fuse(C0Lx, C1Lx, pad=pad)
            CxLx = self.postprocess(CxLx)

        with asection(f"Converting back to original dtype..."):
            if original_dtype is numpy.uint16:
                CxLx = xp.clip(CxLx, 0, None, out=CxLx)
            CxLx = CxLx.astype(dtype=original_dtype, copy=False)

        return CxLx

    @staticmethod
    def _fuse_views_generic(CxL0: xpArray, CxL1: xpArray, bias_axis: int, mode: str, smoothing: int,
                            bias_exponent: int, bias_strength: float, downscale: int = 2) -> xpArray:
        if mode == 'tg':
            fused = fuse_tg_nd(CxL0, CxL1, downscale=downscale, tenengrad_smoothing=smoothing, bias_axis=bias_axis,
                               bias_exponent=bias_exponent, bias_strength=bias_strength)
        elif mode == 'dct':
            fused = fuse_dct_nd(CxL0, CxL1)
        elif mode == 'dft':
            fused = fuse_dft_nd(CxL0, CxL1)
        else:
            raise NotImplementedError
        return fused

    def _fuse_illumination_views(self, CxL0: xpArray, CxL1: xpArray, smoothing: int = 12) -> xpArray:
        return self._fuse_views_generic(CxL0, CxL1, 2, self._fusion, smoothing,
                                        self._fusion_bias_exponent, self._fusion_bias_strength_i)

    def _fuse_detection_views(self, C0Lx: xpArray, C1Lx: xpArray, smoothing: int = 12) -> xpArray:
        return self._fuse_views_generic(C0Lx, C1Lx, 0, self._fusion, smoothing,
                                        self._fusion_bias_exponent, self._fusion_bias_strength_d)

    def compute_registration(self, C0Lx: xpArray, C1Lx: xpArray, mode: str, edge_filter: bool,
                             crop_factor_along_z: float) -> None:
        C0Lx = Backend.to_backend(C0Lx)
        C1Lx = Backend.to_backend(C1Lx)

        depth = C0Lx.shape[0]
        crop = int(depth * crop_factor_along_z)
        C0Lx_c = C0Lx[crop:-crop]
        C1Lx_c = C1Lx[crop:-crop]

        if mode == 'projection':
            self._registration_model = register_translation_proj_nd(C0Lx_c, C1Lx_c, edge_filter=edge_filter)
        elif mode == 'full':
            self._registration_model = register_translation_nd(C0Lx_c, C1Lx_c, edge_filter=edge_filter)
        else:
            raise NotImplementedError

        aprint(f"Applying registration model: {self._registration_model},"
               f"overall confidence: {self._registration_model.overall_confidence()}")
    
    @staticmethod
    def validate_views(views: Dict[str, xpArray]) -> Dict[str, xpArray]:
        if len(views) == 4:
            return views
        assert len(views) == 2, f'Found only {views.keys()}, expected 2 or 4 views.'
        warnings.warn(f"Only two views found. Fusing according to channels' suffixes {views.keys()}")

        if all('C1' in k for k in views):
            warnings.warn("Swapping names from C1 to C0")
            return {k.replace('C1', 'C0'): v for k, v in views.items()}

        if all('L1' in k for k in views):
            warnings.warn("Swapping names from L1 to L0")
            return {k.replace('L1', 'L0'): v for k, v in views.items()}
        
        return views

def summary_from_simview_models(models: List[SimViewFusion]) -> pd.DataFrame:
    df = []
    for m in models:
        current = {}
        for i, eq_ratio in enumerate(m._equalisation_ratios):
            if eq_ratio is not None:
                current[f'eq_ratio_{i}'] = eq_ratio

        if m.registration_model is not None:
            for i, v in enumerate(m.registration_model.shift_vector):
                current[f'shift_{i}'] = v
            current[f'confidence'] = m.registration_model.overall_confidence()

    return pd.DataFrame(df)



import numpy as np
from typing import List, Any, Optional
from arbol.arbol import asection

from dexp.utils import xpArray
from dexp.processing.registration.model import RegistrationModel


class BaseFusion:
    def __init__(self,
                 registration_model: Optional[RegistrationModel],
                 equalise: bool,
                 equalisation_ratios: List[Optional[float]],
                 zero_level: float,
                 clip_too_high: int,
                 fusion: str,
                 dehaze_before_fusion: bool,
                 dehaze_size: int,
                 dehaze_correct_max_level: bool,
                 dark_denoise_threshold: int,
                 dark_denoise_size: int,
                 butterworth_filter_cutoff: float,
                 internal_dtype: np.dtype,
                 ):

        self._registration_model = registration_model
        self._equalise = equalise
        self._equalisation_ratios = equalisation_ratios.copy()
        self._zero_level = zero_level
        self._clip_too_high = clip_too_high
        self._fusion = fusion
        self._dehaze_before_fusion = dehaze_before_fusion
        self._dehaze_size = dehaze_size
        self._dehaze_correct_max_level = dehaze_correct_max_level
        self._dark_denoise_threshold = dark_denoise_threshold
        self._dark_denoise_size = dark_denoise_size
        self._butterworth_filter_cutoff = butterworth_filter_cutoff
        self._internal_dtype = internal_dtype

    @staticmethod
    def _match_input(a: xpArray, b: xpArray) -> None:
        if a.shape != b.shape:
            raise ValueError('The views must have the same shape')
        if a.dtype != b.dtype:
            raise ValueError('The views must have the same dtype')

    def preprocess(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def postprocess(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def fuse(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def compute_registration(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def __call__(self, *args, **kwargs) -> Any:
        pass

    @property
    def registration_model(self) -> RegistrationModel:
        return self._registration_model


from dexp.optics.psf.standard_psfs import nikon16x08na
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.deskew.classic_deskew import classic_deskew_dimensionless
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.utils.timeit import timeit


def demo_classic_deskew_numpy():
    with NumpyBackend():
        _classic_deskew(length=64)


def demo_classic_deskew_cupy():
    try:
        with CupyBackend():
            _classic_deskew()
            return True
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")
        return False


def _classic_deskew(length: int = 128,
                    zoom: int = 4,
                    display: bool = True):

    _deskew(1, length, zoom, display)
    _deskew(0.5, length, zoom, display)


def _deskew(shift, length, zoom, display):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()
    with timeit("generate demo image"):
        # generate nuclei image:
        _, _, image = generate_nuclei_background_data(add_noise=False,
                                                      length_xy=length,
                                                      length_z_factor=1,
                                                      independent_haze=True,
                                                      sphere=True,
                                                      zoom=zoom,
                                                      add_offset=False,
                                                      background_stength=0.07,
                                                      dtype=xp.float32)

        # Pad:
        pad_width = ((int(shift * zoom * length // 2), int(shift * zoom * length // 2)), (0, 0), (0, 0),)
        image = xp.pad(image, pad_width=pad_width)

        # Add blur:
        psf = nikon16x08na()
        psf = psf.astype(dtype=image.dtype, copy=False)
        image = fft_convolve(image, psf)

        # apply skew:
        matrix = xp.asarray([[1, shift, 0], [0, 1, 0], [0, 0, 1]])
        offset = 0 * xp.asarray([image.shape[0] // 2, 0, 0])
        # matrix = xp.linalg.inv(matrix)
        skewed = sp.ndimage.affine_transform(image, matrix, offset=offset)

        # Add noise and clip
        # skewed += xp.random.uniform(-1, 1)
        skewed = xp.clip(skewed, a_min=0, a_max=None)

        # cast to uint16:
        skewed = skewed.astype(dtype=xp.uint16)
    with timeit("deskew image"):
        # apply deskewing:
        deskewed = classic_deskew_dimensionless(skewed,
                                                depth_axis=0,
                                                lateral_axis=1,
                                                shift=shift,
                                                )
    if display:
        from napari import Viewer
        import napari
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer(ndisplay=3)
        viewer.add_image(_c(image), name='image', colormap='bop orange', blending='additive', rendering='attenuated_mip', attenuation=0.01)
        viewer.add_image(_c(skewed), name='skewed_image', colormap='bop blue', blending='additive', visible=False, rendering='attenuated_mip', attenuation=0.01)
        viewer.add_image(_c(deskewed), name='deskewed_image', colormap='bop purple', blending='additive', rendering='attenuated_mip', attenuation=0.01)

        napari.run()
    # compute mean absolute errors:
    error_skewed = xp.mean(xp.absolute(image - skewed))
    error_deskewed = xp.mean(xp.absolute(image - deskewed))
    print(f"error_skewed = {error_skewed}")
    print(f"error_deskewed = {error_deskewed}")
    # Asserts that check if things behave as expected:
    assert error_deskewed < error_skewed
    assert error_skewed > 10
    assert error_deskewed < 2


if __name__ == "__main__":
    if not demo_classic_deskew_cupy():
        demo_classic_deskew_numpy()




import click
from arbol.arbol import aprint, asection

from dexp.cli.defaults import _default_workers_backend
from dexp.cli.parsing import _parse_channels, _parse_slicing, _parse_devices
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.register import dataset_register


@click.command()
@click.argument('input_paths', nargs=-1, required=True)
@click.option('--out-model-path', '-o', default='registration_models.txt', show_default=True)
@click.option('--channels', '-c', default=None, help='list of channels for the view in standard order for the microscope type (C0L0, C0L1, C1L0, C1L1,...)')
@click.option('--slicing', '-s', default=None, help='dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ')
@click.option('--microscope', '-m', type=str, default='simview', help='Microscope objective to use for computing psf, can be: simview or mvsols', show_default=True)
@click.option('--equalise/--no-equalise', '-eq/-neq', default=True, help='Equalise intensity of views before fusion, or not.', show_default=True)
@click.option('--zero-level', '-zl', type=int, default=0, help="‘zero-level’ i.e. the pixel values in the restoration (to be substracted)", show_default=True)
@click.option('--clip-high', '-ch', type=int, default=0, help='Clips voxel values above the given value, if zero no clipping is done', show_default=True)
@click.option('--fusion', '-f', type=str, default='tg', help="Fusion mode, can be: ‘tg’ or ‘dct’.  ", show_default=True)
@click.option('--fusion_bias_strength', '-fbs', type=float, default=0.5, help='Fusion bias strength for illumination', show_default=True)
@click.option('--dehaze_size', '-dhs', type=int, default=65, help='Filter size (scale) for dehazing the final regsitered and fused image to reduce effect of scattered and out-of-focus light. Set to zero to deactivate.', show_default=True)
@click.option('--edge-filter', '-ef', is_flag=True, help='Use this flag to apply an edge filter to help registration.', show_default=True)
@click.option('--max-proj/--no-max-proj', '-mp/-nmp', type=bool, default=True, help='Registers using only the maximum intensity projection from each stack.', show_default=True)
@click.option('--workers', '-k', type=int, default=-1, help='Number of worker threads to spawn, if -1 then num workers = num devices', show_default=True)
@click.option('--workers-backend', '-wkb', type=str, default=_default_workers_backend, help='What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ', show_default=True)
@click.option('--devices', '-d', type=str, default='0', help='Sets the CUDA devices id, e.g. 0,1,2 or ‘all’', show_default=True)
def register(input_paths,
             out_model_path,
             channels,
             slicing,
             microscope,
             equalise,
             zero_level,
             clip_high,
             fusion,
             fusion_bias_strength,
             dehaze_size,
             edge_filter,
             max_proj,
             workers,
             workers_backend,
             devices,
             ):
    """
    Computes registration model for fusing.
    """
    input_dataset, input_paths = glob_datasets(input_paths)
    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)
    devices = _parse_devices(devices)

    with asection(f"Fusing dataset: {input_paths}, saving it at: {out_model_path}, for channels: {channels}, slicing: {slicing} "):
        aprint(f"Microscope type: {microscope}, fusion type: {fusion}")
        aprint(f"Devices used: {devices}, workers: {workers} ")
        dataset_register(dataset=input_dataset,
                         model_path=out_model_path,
                         channels=channels,
                         slicing=slicing,
                         microscope=microscope,
                         equalise=equalise,
                         zero_level=zero_level,
                         clip_too_high=clip_high,
                         fusion=fusion,
                         fusion_bias_strength_i=fusion_bias_strength,
                         dehaze_size=dehaze_size,
                         registration_edge_filter=edge_filter,
                         max_proj=max_proj,
                         workers=workers,
                         workers_backend=workers_backend,
                         devices=devices)



import math
import tempfile
from os.path import join

from arbol import aprint, asection

from dexp.datasets.operations.deskew import dataset_deskew
from dexp.datasets.zarr_dataset import ZDataset
from dexp.optics.psf.standard_psfs import nikon16x08na
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data

# Input dataset
# Please provide path to dataset fro demoing here:
input_path = '/mnt/raid0/pisces_datasets/test.zarr'


def demo_deskew_numpy():
    with NumpyBackend():
        _demo_deskew()


def demo_deskew_cupy():
    try:
        with CupyBackend():
            _demo_deskew(length=128, zoom=4)
            return True
    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")
        return False


def _demo_deskew(length=96,
                 zoom=1,
                 shift=1,
                 display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()


    with asection("Deskew..."):
        with tempfile.TemporaryDirectory() as tmpdir:
            aprint('created temporary directory', tmpdir)

            # Open input dataset:
            dataset = ZDataset(path=input_path, mode='r')

            #Input array:
            input_array = dataset.get_array('v0c0')

            with asection("Deskew yang..."):
                # output path:
                output_path = join(tmpdir, 'deskewed_yang.zarr')

                # deskew:
                dataset_deskew(dataset=dataset,
                               dest_path=output_path,
                               channels=('v0c0',),
                               flips=(True,),
                               slicing=(slice(0, 1),),
                               mode='yang')

                # read result
                deskewed_yang_dataset = ZDataset(path=output_path, mode='r')
                deskewed_yang_array = deskewed_yang_dataset.get_array('v0c0')

                assert deskewed_yang_array.shape[0] == 1

            with asection("Deskew classic..."):
                # output path:
                output_path = join(tmpdir, 'deskewed_classic.zarr')

                # deskew:
                dataset_deskew(dataset=dataset,
                               dest_path=output_path,
                               channels=('v0c0',),
                               flips=(True,),
                               slicing=(slice(0, 1),),
                               mode='classic',
                               padding=True)

                # read result
                deskewed_classic_dataset = ZDataset(path=output_path, mode='r')
                deskewed_classic_array = deskewed_classic_dataset.get_array('v0c0')

                assert deskewed_classic_array.shape[0] == 1

                deskewed_classic_array = xp.rot90(Backend.to_backend(deskewed_classic_array[0]), 1, axes=(0, 1))
                deskewed_classic_array = xp.rot90(Backend.to_backend(deskewed_classic_array), 1, axes=(1, 2))

            if display:

                def _c(array):
                    return Backend.to_numpy(array)

                import napari
                viewer = napari.Viewer(ndisplay=3)
                viewer.add_image(_c(input_array), name='input_array')
                viewer.add_image(_c(deskewed_yang_array), name='deskewed_yang_array')
                viewer.add_image(_c(deskewed_classic_array), name='deskewed_classic_array')
                viewer.grid.enabled = True
                napari.run()


if __name__ == "__main__":
    if not demo_deskew_cupy():
        demo_deskew_numpy()



import numpy

from dexp.optics.psf.standard_psfs import nikon16x08na
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.processing.deconvolution.admm_deconvolution import admm_deconvolution
from dexp.processing.deconvolution.lr_deconvolution import lucy_richardson_deconvolution
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.processing.deconvolution.blind_deconvolution import blind_deconvolution
from dexp.utils.timeit import timeit

from functools import partial


def demo_blind_deconvolution_numpy():
    with NumpyBackend():
        _demo_blind_deconvolution()


def demo_blind_deconvolution_cupy():
    try:
        with CupyBackend():
            _demo_blind_deconvolution()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_blind_deconvolution(length_xy=128):
    xp = Backend.get_xp_module()

    with timeit("generate data"):
        image_gt, background, image = generate_nuclei_background_data(add_noise=False,
                                                                      length_xy=length_xy,
                                                                      length_z_factor=1,
                                                                      background_stength=0,
                                                                      add_offset=False)

    psf = nikon16x08na(xy_size=35, z_size=35) # , dxy=0.200, dz=0.94)
    psf = psf.astype(dtype=numpy.float16)
    image = image.astype(dtype=numpy.float16)

    blurry = fft_convolve(image, psf)
    blurry = blurry - blurry.min()
    blurry = blurry / blurry.max()
    noisy = blurry
    noisy = noisy + 0.05 * xp.random.uniform(size=blurry.shape)

    iterations = 10
    # normalizing (required by admm)
    noisy = noisy - noisy.min()
    noisy = noisy / noisy.max()

    # NOTE: could not make it work admm deconvolution (yet)
    # deconv = partial(admm_deconvolution, rho=10, gamma=0.01, iterations=iterations, derivative=2)
    deconv = partial(lucy_richardson_deconvolution, num_iterations=50, padding=psf.shape[2] // 2)
    params = dict(wl=561, na=0.8, ni=1.33, res=485, zres=2000)

    psf = (10000 * psf).astype(int)
    wrong_psf = psf + (psf * 0.1) * numpy.random.uniform(size=psf.shape)
    wrong_psf = psf.astype(int)

    with timeit("Computing blind deconvolution"):
        deconvolved = blind_deconvolution(noisy, wrong_psf, deconv, params, 5)

    def _c(array):
        return Backend.to_numpy(array)

    import napari
    viewer = napari.Viewer()
    viewer.add_image(_c(image), name='image')
    viewer.add_image(_c(blurry), name='blurry')
    viewer.add_image(_c(psf), name='psf')
    viewer.add_image(_c(noisy), name='noisy')
    viewer.add_image(_c(deconvolved), name='sparse deconv')

    napari.run()



if __name__ == "__main__":
    demo_blind_deconvolution_cupy()
    demo_blind_deconvolution_numpy()



import random
import tempfile
from os.path import join

from arbol import aprint, asection

from dexp.datasets.operations.copy import dataset_copy
from dexp.datasets.zarr_dataset import ZDataset
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data


def demo_copy_numpy():
    with NumpyBackend():
        _demo_copy()


def demo_copy_cupy():
    try:
        with CupyBackend():
            _demo_copy(length_xy=128, zoom=2)
            return True
    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")
        return False


def _demo_copy(length_xy=96,
               zoom=1,
               n=16,
               display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    # generate nuclei image:
    _, _, image = generate_nuclei_background_data(add_noise=False,
                                                  length_xy=length_xy,
                                                  length_z_factor=1,
                                                  independent_haze=True,
                                                  sphere=True,
                                                  zoom=zoom,
                                                  dtype=xp.float32)

    with asection("prepare simulated timelapse:"):
        # move to backend:
        image = Backend.to_backend(image)

        # generate reference 'ground truth' timelapse
        images = (image.copy() for _ in range(n))

        # modify each image:
        images = (sp.ndimage.shift(image, shift=(random.uniform(-1.0, 1.0), random.uniform(-1.0, 1.0), random.uniform(-1.0, 1.0))) for image in images)
        images = (image + random.uniform(-10, 10) for image in images)

        # turn into array:
        images = xp.stack(images)

    with tempfile.TemporaryDirectory() as tmpdir:
        aprint('created temporary directory', tmpdir)

        with asection("Prepare dataset..."):
            input_path = join(tmpdir, 'dataset.zarr')
            dataset = ZDataset(path=input_path,
                               mode='w',
                               store='dir')

            dataset.add_channel(name='channel',
                                shape=images.shape,
                                chunks=(1, 64, 64, 64),
                                dtype=images.dtype)

            dataset.write_array(channel='channel',
                                array=Backend.to_numpy(images))

            source_array = dataset.get_array('channel')

        with asection("Copy..."):
            # output_folder:
            output_path = join(tmpdir, 'copy.zarr')


            # Do actual copy:
            dataset_copy(dataset=dataset,
                         dest_path=output_path,
                         channels=('channel',),
                         slicing=(slice(0, 4), ...),
                         workers=4,
                         workersbackend='threading')

            copied_dataset = ZDataset(path=output_path, mode='a')
            copied_array = copied_dataset.get_array('channel')

            assert copied_array.shape[0] == 4
            assert copied_array.shape[1:] == source_array.shape[1:]


        if display:

            def _c(array):
                return Backend.to_numpy(array)

            import napari
            viewer = napari.Viewer(ndisplay=3)
            viewer.add_image(_c(source_array), name='source_array')
            viewer.add_image(_c(copied_array), name='copied_array')
            viewer.grid.enabled = True
            napari.run()


if __name__ == "__main__":
    if not demo_copy_cupy():
        demo_copy_numpy()



import numpy as np
from typing import Tuple, List


def slice_from_shape(shape: Tuple[int], slicing: slice) -> Tuple[Tuple[int], slice, List[int]]:
    time_points = list(range(shape[0]))
    if slicing is not None:
        if isinstance(slicing, tuple):
            time_points = time_points[slicing[0]]
            volume_slicing = slicing[1:]
            new_shape = (len(time_points),) + np.empty(shape=shape[1:], dtype=bool)[slicing[1:]].shape
        else:
            time_points = time_points[slicing]
            volume_slicing = ...
            new_shape = (len(time_points),) + shape[1:]
    else:
        volume_slicing = ...
        new_shape = shape
    return new_shape, volume_slicing, time_points



from typing import Optional, List, Sequence

from arbol.arbol import aprint
from arbol.arbol import asection
from joblib import Parallel, delayed
from zarr.errors import ContainsArrayError, ContainsGroupError

from dexp.datasets.base_dataset import BaseDataset
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.best_backend import BestBackend
from dexp.processing.deskew.classic_deskew import classic_deskew
from dexp.processing.deskew.yang_deskew import yang_deskew


def dataset_deskew(dataset: BaseDataset,
                   dest_path: str,
                   channels: Sequence[str],
                   slicing,
                   dx: Optional[float] = None,
                   dz: Optional[float] = None,
                   angle: Optional[float] = None,
                   flips: Optional[Sequence[bool]] = None,
                   camera_orientation: int = 0,
                   depth_axis: int = 0,
                   lateral_axis: int = 1,
                   mode: str = 'classic',
                   padding: bool = True,
                   store: str = 'dir',
                   compression: str = 'zstd',
                   compression_level: int = 3,
                   overwrite: bool = False,
                   workers: int = 1,
                   workersbackend: str = 'threading',
                   devices: Optional[List[int]] = None,
                   check: bool = True,
                   stop_at_exception=True):

    # Collect arrays for selected channels:
    arrays = tuple(dataset.get_array(channel, per_z_slice=False, wrap_with_dask=True) for channel in channels)

    # Notify of selected channels and corresponding properties:
    with asection(f"Channels:"):
        for view, channel in zip(arrays, channels):
            aprint(f"Channel: {channel} of shape: {view.shape} and dtype: {view.dtype}")

    # Slicing:
    if slicing is not None:
        aprint(f"Slicing with: {slicing}")
        arrays = tuple(view[slicing] for view in arrays)

    # Default flipping:
    if flips is None:
        flips = (False,) * len(arrays)

    # Default devices:
    if devices is None:
        devices = [0]

    # We allocate last minute once we know the shape...
    from dexp.datasets.zarr_dataset import ZDataset
    zarr_mode = 'w' + ('' if overwrite else '-')
    dest_dataset = ZDataset(dest_path, zarr_mode, store, parent=dataset)

    # Metadata for deskewing:
    metadata = dataset.get_metadata()
    aprint(f"Dataset metadata: {metadata}")
    if dx is None and 'res' in metadata:
        dx = float(metadata['res'])
    if dz is None and 'dz' in metadata:
        dz = float(metadata['dz']) if dz is None else dz
    if angle is None and 'angle' in metadata:
        angle = float(metadata['angle']) if angle is None else angle

    aprint(f"Deskew parameters: dx={dx}, dz={dz}, angle={angle}")

    # Iterate through channels::
    for array, channel, flip in zip(arrays, channels, flips):

        # shape and dtype of views to deskew:
        shape = array[0].shape
        nb_timepoints = array.shape[0]

        def process(channel, tp, device):
            try:

                with asection(f"Loading channel {channel} for time point {tp}"):
                    array_tp = array[tp].compute()

                with BestBackend(device, exclusive=True, enable_unified_memory=True):

                    if 'yang' in mode:
                        deskewed_view_tp = yang_deskew(image=array_tp,
                                                        depth_axis=depth_axis,
                                                        lateral_axis=lateral_axis,
                                                        flip_depth_axis=flip,
                                                        dx=dx,
                                                        dz=dz,
                                                        angle=angle,
                                                        camera_orientation=camera_orientation)
                    elif 'classic' in mode:
                        deskewed_view_tp = classic_deskew(image=array_tp,
                                                           depth_axis=depth_axis,
                                                           lateral_axis=lateral_axis,
                                                           flip_depth_axis=flip,
                                                           dx=dx,
                                                           dz=dz,
                                                           angle=angle,
                                                           camera_orientation=camera_orientation,
                                                           padding=padding)
                    else:
                        raise ValueError(f"Deskew mode: {mode} not supported.")

                    with asection(f"Moving array from backend to numpy."):
                        deskewed_view_tp = Backend.to_numpy(deskewed_view_tp, dtype=array.dtype, force_copy=False)

                if channel not in dest_dataset.channels():
                    try:
                        dest_dataset.add_channel(channel,
                                                 shape=(array.shape[0],)+deskewed_view_tp.shape,
                                                 dtype=deskewed_view_tp.dtype,
                                                 codec=compression,
                                                 clevel=compression_level)
                    except (ContainsArrayError, ContainsGroupError):
                        aprint(f"Other thread/process created channel before... ")

                with asection(f"Saving fused stack for time point {tp}, shape:{deskewed_view_tp.shape}, dtype:{deskewed_view_tp.dtype}"):
                    dest_dataset.write_stack(channel=channel,
                                             time_point=tp,
                                             stack_array=deskewed_view_tp)

                aprint(f"Done processing time point: {tp} .")

            except Exception as error:
                aprint(error)
                aprint(f"Error occurred while deskewing time point {tp} !")
                import traceback
                traceback.print_exc()

                if stop_at_exception:
                    raise error

        if workers == -1:
            workers = len(devices)
        aprint(f"Number of workers: {workers}")

        if workers > 1:
            Parallel(n_jobs=workers, backend=workersbackend)(delayed(process)(channel, tp, devices[tp % len(devices)]) for tp in range(0, shape[0]))
        else:
            for tp in range(0, nb_timepoints):
                process(channel, tp, devices[0])

    # Dataset info:
    aprint(dest_dataset.info())

    # Check dataset integrity:
    if check:
        dest_dataset.check_integrity()

    # close destination dataset:
    dest_dataset.close()



import math
import tempfile
from os.path import join

from arbol import aprint, asection

from dexp.datasets.operations.deskew import dataset_deskew
from dexp.datasets.zarr_dataset import ZDataset
from dexp.optics.psf.standard_psfs import nikon16x08na
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data


def demo_deskew_numpy():
    with NumpyBackend():
        _demo_deskew()


def demo_deskew_cupy():
    try:
        with CupyBackend():
            _demo_deskew(length=128, zoom=4)
            return True
    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")
        return False


def _demo_deskew(length=96,
                 zoom=1,
                 shift=1,
                 display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    with asection("prepare simulated timelapse:"):
        # generate nuclei image:
        _, _, image = generate_nuclei_background_data(add_noise=False,
                                                      length_xy=length,
                                                      length_z_factor=1,
                                                      independent_haze=True,
                                                      sphere=True,
                                                      zoom=zoom,
                                                      add_offset=False,
                                                      background_stength=0.07,
                                                      dtype=xp.float32)

        # Pad:
        pad_width = ((int(shift * zoom * length // 2), int(shift * zoom * length // 2)), (0, 0), (0, 0),)
        image = xp.pad(image, pad_width=pad_width)

        # Add blur:
        psf = nikon16x08na()
        psf = psf.astype(dtype=image.dtype, copy=False)
        image = fft_convolve(image, psf)

        # apply skew:
        angle = 45
        matrix = xp.asarray([[math.cos(angle * math.pi / 180), math.sin(angle * math.pi / 180), 0], [0, 1, 0], [0, 0, 1]])
        offset = 0 * xp.asarray([image.shape[0] // 2, 0, 0])
        # matrix = xp.linalg.inv(matrix)
        skewed = sp.ndimage.affine_transform(image, matrix, offset=offset)

        # Add noise and clip
        # skewed += xp.random.uniform(-1, 1)
        skewed = xp.clip(skewed, a_min=0, a_max=None)

        # cast to uint16:
        skewed = skewed.astype(dtype=xp.uint16)

        # Single timepoint:
        skewed = skewed[xp.newaxis, ...]


    with tempfile.TemporaryDirectory() as tmpdir:
        aprint('created temporary directory', tmpdir)

        with asection("Prepare dataset..."):
            input_path = join(tmpdir, 'dataset.zarr')
            dataset = ZDataset(path=input_path,
                               mode='w',
                               store='dir')

            dataset.add_channel(name='channel',
                                shape=skewed.shape,
                                chunks=(1, 64, 64, 64),
                                dtype=skewed.dtype)

            dataset.write_array(channel='channel',
                                array=Backend.to_numpy(skewed))

            source_array = dataset.get_array('channel')

        with asection("Deskew..."):
            # output_folder:
            output_path = join(tmpdir, 'deskewed.zarr')

            # deskew:
            dataset_deskew(dataset=dataset,
                           dest_path=output_path,
                           channels=('channel',),
                           slicing=(slice(0, 1),),
                           dx=1,
                           dz=1,
                           angle=angle)

            deskew_dataset = ZDataset(path=output_path, mode='a')
            deskew_array = deskew_dataset.get_array('channel')

            assert deskew_array.shape[0] == 1


        if display:

            def _c(array):
                return Backend.to_numpy(array)

            import napari
            viewer = napari.Viewer(ndisplay=3)
            viewer.add_image(_c(source_array), name='source_array')
            viewer.add_image(_c(deskew_array), name='deconv_array')
            viewer.grid.enabled = True
            napari.run()


if __name__ == "__main__":
    if not demo_deskew_cupy():
        demo_deskew_numpy()



from arbol import aprint

from dexp.datasets.operations.demo.demo_deconv import _demo_deconv
from dexp.datasets.operations.demo.demo_deskew import _demo_deskew
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.deskew.demo.demo_classic_deskew import _classic_deskew
from dexp.processing.deskew.demo.demo_yang_deskew import _yang_deskew


def test_yang_deskew_numpy():
    with NumpyBackend():
        _yang_deskew(length=48, display=False)

def test_yang_deskew_cupy():
    try:
        with CupyBackend():
            _yang_deskew(length=48, display=False)

    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")





import os
import shutil
import sys
import re
from pathlib import Path
from os.path import isfile, isdir, exists, join
from typing import Tuple, Sequence, Any, Union, Optional, List

import dask
import numpy
import zarr
from arbol.arbol import aprint, asection
from zarr import open_group, convenience, CopyError, Blosc, Group
from ome_zarr.format import CurrentFormat

from dexp.datasets.base_dataset import BaseDataset
from dexp.datasets.ome_dataset import default_omero_metadata
from dexp.processing.backends.backend import Backend

from dexp.utils.config import config_blosc


class ZDataset(BaseDataset):
    def __init__(self, path: str, mode: str = 'r', store: str = None, parent: Optional[BaseDataset] = None):
        """Instantiates a Zarr dataset (and opens it)

        Parameters
        ----------
        path : path to zarr storage (directory or zip).
        mode : Access mode:
            'r' means read only (must exist);
            'r+' means read/write (must exist);
            'a' means read/write (create if doesn't exist);
            'w' means create (overwrite if exists);
            'w-' means create (fail if exists).
        store : type of store, can be 'dir', 'ndir', or 'zip'

        Returns
        -------
        Zarr dataset
        """
        config_blosc()

        super().__init__(dask_backed=False)

        self._path = path
        self._store = None
        self._root_group = None
        self._arrays = {}
        self._projections = {}

        # Open remote store:
        if 'http' in path:
            aprint(f"Opening a remote store at: {path}")
            from fsspec import get_mapper
            self.store = get_mapper(path)
            self._root_group = zarr.open(self.store, mode=mode)
            self._initialise_existing()
            return

        # Correct path to adhere to convention:
        if 'a' in mode or 'w' in mode:
            if path.endswith('.zarr.zip') or store == 'zip':
                path = path + '.zip' if path.endswith('.zarr') else path
                path = path if path.endswith('.zarr.zip') else path + '.zarr.zip'
            elif path.endswith('.nested.zarr') or path.endswith('.nested.zarr/') or store == 'ndir':
                path = path if path.endswith('.nested.zarr') else path + '.nested.zarr'
            elif path.endswith('.zarr') or path.endswith('.zarr/') or store == 'dir':
                path = path if path.endswith('.zarr') else path + '.zarr'

        # if exists and overwrite then delete!
        if exists(path) and mode == 'w-':
            raise ValueError(f"Storage '{path}' already exists, add option '-w' to force overwrite!")
        elif exists(path) and mode == 'w':
            aprint(f"Deleting '{path}' for overwrite!")
            if isdir(path):
                # This is a very dangerous operation, let's double check that the folder really holds a zarr dataset:
                _zgroup_file = join(path, '.zgroup')
                _zarray_file = join(path, '.zarray')
                # We check that either of these two hiddwn files are present, and if '.zarr' is part of the name:
                if ('.zarr' in path) and (exists(_zgroup_file) or exists(_zarray_file)):
                    shutil.rmtree(path, ignore_errors=True)
                else:
                    raise ValueError("Specified path does not seem to be a zarr dataset, deletion for overwrite not performed out of abundance of caution, check path!")

            elif isfile(path):
                os.remove(path)

        if exists(path):
            aprint(f"Opening existing Zarr storage: '{path}' with read/write mode: '{mode}' and store type: '{store}'")
            if isfile(path) and (path.endswith('.zarr.zip') or store == 'zip'):
                aprint(f"Opening as ZIP store")
                self._store = zarr.storage.ZipStore(path)
            elif isdir(path) and ((path.endswith('.nested.zarr') or path.endswith('.nested.zarr/') or store == 'ndir')):
                aprint(f"Opening as Nested Directory store")
                self._store = zarr.storage.NestedDirectoryStore(path)
            elif isdir(path) and ((path.endswith('.zarr') or path.endswith('.zarr/') or store == 'dir')):
                aprint(f"Opening as Directory store")
                self._store = zarr.storage.DirectoryStore(path)

            aprint(f"Opening with mode: {mode}")
            self._root_group = open_group(self._store, mode=mode)
            self._initialise_existing()
        elif 'a' in mode or 'w' in mode:
            aprint(f"Creating Zarr storage: '{path}' with read/write mode: '{mode}' and store type: '{store}'")
            if store is None:
                store = 'dir'
            try:
                if path.endswith('.zarr.zip') or store == 'zip':
                    aprint(f"Opening as ZIP store")
                    self._store = zarr.storage.ZipStore(path)
                elif path.endswith('.nested.zarr') or path.endswith('.nested.zarr/') or store == 'ndir':
                    aprint(f"Opening as Nested Directory store")
                    self._store = zarr.storage.NestedDirectoryStore(path)
                elif path.endswith('.zarr') or path.endswith('.zarr/') or store == 'dir':
                    aprint(f"Opening as Directory store")
                    self._store = zarr.storage.DirectoryStore(path)
                else:
                    aprint(f'Cannot open {path}, needs to be a zarr directory (directory that ends with `.zarr` or `.nested.zarr` for nested folders), or a zipped zarr file (file that ends with `.zarr.zip`)')

                self._root_group = zarr.convenience.open(self._store, mode=mode)

            except Exception as e:
                raise ValueError(f"Problem: can't create target file/directory, most likely the target dataset already exists or path incorrect: {path}")
        else:
            raise ValueError(f"Invalid read/write mode or invalid path: {path} (check path!)")

        # updating metadata
        if parent is not None:
            metadata = parent.get_metadata()
            metadata.pop('cli_history', None)  # avoiding adding it twice
            self.append_metadata(metadata)

        if mode in ('a', 'w', 'w-'):
            self.append_cli_history(parent if isinstance(parent, ZDataset) else None)

    def _initialise_existing(self):
        self._channels = [channel for channel, _ in self._root_group.groups()]

        aprint(f"Exploring Zarr hierarchy...")
        for channel, channel_group in self._root_group.groups():
            aprint(f"Found channel: {channel}")

            channel_items = channel_group.items()

            for item_name, array in channel_items:
                aprint(f"Found array: {item_name}")

                if item_name == channel or item_name == 'fused':
                    # print(f'Opening array at {path}:{channel}/{item_name} ')
                    self._arrays[channel] = array
                    # self._arrays[channel] = from_zarr(path, component=f"{channel}/{item_name}")
                elif (item_name.startswith(channel) or item_name.startswith('fused')) and '_projection_' in item_name:
                    self._projections[item_name] = array

    def _get_group_for_channel(self, channel: str) -> Union[None, Sequence[Group]]:
        groups = [g for c, g in self._root_group.groups() if c == channel]
        if len(groups) == 0:
            return None
        else:
            return groups[0]

    @staticmethod
    def _default_chunks(shape: Tuple[int], dtype: Union[str, numpy.dtype], max_size: int = 2147483647) -> Tuple[int]:
        chunk = (1, 64, 256, 256)  # testing smaller chunks
        # if not isinstance(dtype, numpy.dtype):
        #     dtype = numpy.dtype(dtype)
        # width = shape[-1]
        # height = shape[-2]
        # depth = min(max_size // (dtype.itemsize * width * height), shape[-3])
        # chunk = (1, depth, height, width)
        return chunk[-len(shape):]

    def close(self):
        # We close the store if it exists, i.e. if we have been writing to the dataset
        if self._store is not None:
            try:
                self._store.close()
            except AttributeError:
                pass

    def check_integrity(self, channels: Sequence[str] = None) -> bool:
        aprint(f"Checking integrity of zarr storage, might take some time.")
        if channels is None:
            channels = self.channels()
        for channel in channels:
            aprint(f"Checking integrity of channel '{channel}'...")
            array = self.get_array(channel, wrap_with_dask=False)
            if array.nchunks_initialized < array.nchunks:
                aprint(f"WARNING! not all chunks initialised! (dtype={array.dtype})")
                return False
            else:
                aprint(f"Channel '{channel}' seems ok!")
                return True

    def channels(self) -> Sequence[str]:
        return list(self._arrays.keys())

    def nb_timepoints(self, channel: str) -> int:
        return self.get_array(channel).shape[0]

    def shape(self, channel: str) -> Sequence[int]:
        return self._arrays[channel].shape

    def chunks(self, channel: str) -> Sequence[int]:
        return self._arrays[channel].chunks

    def dtype(self, channel: str):
        return self.get_array(channel).dtype

    def info(self, channel: str = None, cli_history: bool = True) -> str:
        info_str = ''
        if channel is not None:
            info_str += f"Channel: '{channel}', nb time points: {self.shape(channel)[0]}, shape: {self.shape(channel)[1:]}"
            info_str += ".\n"
            info_str += str(self._arrays[channel].info)
            return info_str
        else:
            info_str += f"Dataset at location: {self._path} \n"
            info_str += f"Channels: {self.channels()} \n"
            info_str += "Zarr tree: \n"
            info_str += str(self._root_group.tree())
            info_str += ".\n\n"
            info_str += "Arrays: \n"
            for name, array in self._arrays.items():
                info_str += "  │ \n"
                info_str += "  └──" + name +":\n" + str(array.info) + "\n\n"
                info_str += ".\n\n"

        info_str += ".\n\n"
        info_str += f"\nMetadata: \n"
        for key, value in self.get_metadata().items():
            if 'cli_history' not in key:
                info_str += f"\t{key} : {value} \n"

        if cli_history:
            info_str += ".\n\n"
            key = 'cli_history'
            if key in self._root_group.attrs:
                info_str += "\nCommand line history:\n"
                commands_list = self._root_group.attrs[key]
                for command in commands_list[:-1]:
                    info_str += " ├──■ '" + command + "' \n"
                info_str += " └──■ '" + commands_list[-1] + "' \n"

        return info_str

    def get_metadata(self):
        """get the attributes stored in the zarr folder"""
        attrs = {}
        for name in self._root_group.attrs:
            attrs[name] = self._root_group.attrs[name]
        return attrs

    def append_metadata(self, metadata: dict):
        self._root_group.attrs.update(metadata)

    def append_cli_history(self, parent: Optional[BaseDataset]):
        key = 'cli_history'
        cli_history = []
        if parent is not None:
            parent_metadata = parent.get_metadata()
            cli_history = parent_metadata.get(key, [])

        if key in self._root_group.attrs:
            cli_history += self._root_group.attrs[key]

        new_command = os.path.basename(sys.argv[0]) + ' ' + ' '.join(sys.argv[1:])
        cli_history.append(new_command)
        self._root_group.attrs[key] = cli_history

    def _load_tensorstore(self, array: zarr.Array):
        import tensorstore as ts

        metadata = {
            'dtype': array.dtype.str,
            'order': array.order,
            'shape': array.shape,
        }
        ts_spec ={
            'driver': 'zarr',
            'kvstore': {
                'driver': 'file',
                'path': self._path,
            },
            'path': array.path,
            'metadata': metadata,
        }
        return ts.open(ts_spec, create=False, open=True).result()

    def get_array(self, channel: str, per_z_slice: bool = False, wrap_with_dask: bool = False, wrap_with_tensorstore: bool = False):
        assert (wrap_with_dask != wrap_with_tensorstore) or wrap_with_dask == False
        array = self._arrays[channel]
        if wrap_with_dask:
            return dask.array.from_array(array, chunks=array.chunks)
        elif wrap_with_tensorstore:
            return self._load_tensorstore(array)
        return array

    def get_stack(self, channel: str, time_point: int, per_z_slice: bool = False, wrap_with_dask: bool = False):
        stack_array = self.get_array(channel, per_z_slice=per_z_slice, wrap_with_dask=wrap_with_dask)[time_point]
        return stack_array

    def get_projection_array(self, channel: str, axis: int, wrap_with_dask: bool = False) -> Any:
        array = self._projections.get(self._projection_name(channel, axis))
        if array is None:
            return
        return dask.array.from_array(array, chunks=array.chunks) if wrap_with_dask else array

    def _projection_name(self, channel: str, axis: int):
        return f'{channel}_projection_{axis}'

    def write_stack(self, channel: str, time_point: int, stack_array: numpy.ndarray):
        array_in_zarr = self.get_array(channel=channel,
                                       wrap_with_dask=False)
        array_in_zarr[time_point] = stack_array

        for axis in range(stack_array.ndim):
            xp = Backend.get_xp_module()
            projection = xp.max(stack_array, axis=axis)
            projection_in_zarr = self.get_projection_array(channel=channel,
                                                           axis=axis,
                                                           wrap_with_dask=False)
            projection_in_zarr[time_point] = projection

    def write_array(self, channel: str, array: numpy.ndarray):
        array_in_zarr = self.get_array(channel=channel,
                                       wrap_with_dask=False)
        array_in_zarr[...] = array

        for axis in range(array.ndim - 1):
            xp = Backend.get_xp_module()
            projection = xp.max(array, axis=axis + 1)
            projection_in_zarr = self.get_projection_array(channel=channel,
                                                           axis=axis,
                                                           wrap_with_dask=False)
            projection_in_zarr[...] = projection

    def add_channel(self,
                    name: str,
                    shape: Tuple[int, ...],
                    dtype: numpy.dtype,
                    chunks: Sequence[int] = None,
                    enable_projections: bool = True,
                    codec: str = 'zstd',
                    clevel: int = 3,
                    value: Optional[Any] = None) -> Any:
        """Adds a channel to this dataset

        Parameters
        ----------
        name : name of channel.
        shape : shape of correspodning array.
        dtype : dtype of array.
        chunks: chunks shape.
        codec: Compression codec to be used ('zstd', 'blosclz', 'lz4', 'lz4hc', 'zlib' or 'snappy').
        clevel: An integer between 0 and 9 specifying the compression level.

        Returns
        -------
        zarr array


        """
        # check if channel exists:
        if name in self.channels():
            raise ValueError("Channel already exist!")

        if chunks is None:
            chunks = self._default_chunks(shape, dtype)

        aprint(f"chunks={chunks}")

        # Choosing the fill value to the largest value:
        fill_value = self._get_largest_dtype_value(dtype) if value is None else value

        aprint(f"Adding channel: '{name}' of shape: {shape}, chunks:{chunks}, dtype: {dtype}, fill_value: {fill_value}, codec: {codec}, clevel: {clevel} ")
        compressor = Blosc(cname=codec, clevel=clevel, shuffle=Blosc.BITSHUFFLE)
        filters = []

        channel_group = self._root_group.create_group(name)
        array = channel_group.full(name=name,
                                   shape=shape,
                                   dtype=dtype,
                                   chunks=chunks,
                                   filters=filters,
                                   compressor=compressor,
                                   fill_value=fill_value)

        self._arrays[name] = array

        if enable_projections:
            ndim = len(shape) - 1
            for axis in range(ndim):
                proj_name = self._projection_name(name, axis)

                proj_shape = list(shape)
                del proj_shape[1 + axis]
                proj_shape = tuple(proj_shape)

                # chunking along time must be 1 to allow parallelism, but no chunking for each projection (not needed!)
                proj_chunks = (1,) + (None,) * (len(chunks) - 2)

                proj_array = channel_group.full(name=proj_name,
                                                shape=proj_shape,
                                                dtype=dtype,
                                                chunks=proj_chunks,
                                                filters=filters,
                                                compressor=compressor,
                                                fill_value=fill_value)
                self._projections[proj_name] = proj_array

        return array

    def add_channels_to(self,
                        zdataset: Union[str, 'ZDataset'],
                        channels: Sequence[str],
                        rename: Sequence[str],
                        store: str = None,
                        add_projections: bool = True,
                        overwrite: bool = True,
                        ):
        """Adds channels from this zarr dataset into an other possibly existing zarr dataset

        Parameters
        ----------
        path : zarr dataset or path of zarr dataset.
        channels: list or tuple of channels to add
        rename: list or tuple of new names for channels
        store: type of zarr store: 'dir' or 'zip', only usefull if store does not exist yet!
        add_projections: If True the projections are also copied.
        overwrite: overwrite destination (not fully functional for zip stores!)

        """

        if type(zdataset) is str:
            zdataset = ZDataset(zdataset, 'a', store, parent=self)

        root = zdataset._root_group

        aprint(f"Existing channels: {zdataset.channels()}")

        for channel, new_name in zip(channels, rename):
            try:
                array = self.get_array(channel, per_z_slice=False, wrap_with_dask=False)
                source_group = self._get_group_for_channel(channel)
                source_arrays = source_group.items()

                aprint(f"Creating group for channel {channel} of new name {new_name}.")
                if new_name not in root.group_keys():
                    dest_group = root.create_group(new_name)
                else:
                    dest_group = root[new_name]

                aprint(f"Fast copying channel {channel} renamed to {new_name} of shape {array.shape} and dtype {array.dtype} ")

                for name, array in source_arrays:
                    if name in self.channels():
                        aprint(f"Fast copying array {name} to {new_name}")
                        convenience.copy(source=array,
                                         dest=dest_group,
                                         name=new_name,
                                         if_exists='replace' if overwrite else 'raise')

                        if add_projections:
                            ndim = array.ndim - 1
                            for axis in range(ndim):
                                proj_array = self.get_projection_array(channel=channel,
                                                                       axis=axis,
                                                                       wrap_with_dask=False)
                                convenience.copy(source=proj_array,
                                                 dest=dest_group,
                                                 name=self._projection_name(new_name, axis),
                                                 if_exists='replace' if overwrite else 'raise')

            except (CopyError, NotImplementedError):
                aprint(f"Channel already exists, set option '-w' to force overwriting! ")

        zdataset.close()

    def get_resolution(self, channel: Optional[str] = None) -> List[float]:
        """
        Gets pixel resolution.

        Parameters
        ----------
        channel : Channel to obtain the information, if None returns the dataset default.
        """
        axes = ('dt', 'dz', 'dy', 'dx')
        metadata = self.get_metadata()
        if channel is None:
            resolution = [metadata.get(axis, 1.0) for axis in axes]

        else:
            resolution = self.get_resolution()  # gets dataset default
            channel_metadata = metadata.get(channel, {})
            resolution = [
                channel_metadata.get(axis, s)
                for axis, s in zip(axes, resolution)
            ]

        return resolution
    
    def get_translation(self, channel: Optional[str] = None) -> List[float]:
        """
        Gets channel translation.
        """
        axes = ('tz', 'ty', 'tx')
        metadata = self.get_metadata()
        if channel is None:
            translation = [metadata.get(axis, 0) for axis in axes]
        
        else:
            translation = self.get_translation()  # gets default translation
            channel_metadata = metadata.get(channel, {})
            translation = [
                channel_metadata.get(axis, t)
                for axis, t in zip(axes, translation)
            ]
        
        return translation

    def to_bdv_format(self, channel: str, path: Union[str, Path]) -> None:
        import npy2bdv

        with asection('Writing BigDataViewer file'):
            if isinstance(path, str):
                path = Path(path)

            if path.exists():
                raise ValueError(f'Path: {path} exists!')
            
            array = self.get_array(channel)
            # ignoring time and reversing to fiji ordering
            resolution = self.get_resolution(channel)[1:][::-1]
            writer = npy2bdv.BdvWriter(str(path))

            for t in range(array.shape[0]):
                writer.append_view(array[t], time=t, calibration=resolution)
                aprint(f'Saved time point {t}')

            writer.write_xml()
            writer.close()
    
    def first_uninitialized_time_point(self, channel: str) -> int:
        """
        Returns the index of the first uninitialized time point or the last time point if it is fully initialized
        """
        array = self.get_array(channel)
        prog = re.compile(r'\.'.join([r'\d+'] * min(1, array.ndim)))
        initialized = set(int(k.split('.', 1)[0]) for k in zarr.storage.listdir(array.chunk_store, array._path) if prog.match(k))
        return max(initialized)

    def to_ome_zarr(self, path: str, chunks: Optional[Sequence[int]] = None,
                    force_dtype: Optional[int] = None):
        # TODO add multiscale support

        ch = self.channels()[0]
        dexp_shape = self.shape(ch)

        dtype = force_dtype if force_dtype is not None else self.dtype(ch)

        for _ch in self.channels():
            if dexp_shape != self.shape(_ch):
                raise ValueError(f'Channels {ch} and {_ch} have different '
                                    f'shapes ({dexp_shape} and {self.shape(_ch)} '
                                    'could not convert to ome-zarr.')
            if dtype != self.dtype(_ch) and force_dtype is None:
                raise ValueError(f'Channels {ch} and {_ch} have different '
                                    f'dtypes ({dtype} and {self.dtype(_ch)} '
                                    'could not convert to ome-zarr.')
                 
        if chunks is None:
            chunks = (1,) + self._default_chunks(dexp_shape, dtype=dtype)
        
        if len(chunks) != 5:
            raise ValueError(f'Chunks must be 5-dimensional. Found {chunks}.')

        ome_zarr_shape = (dexp_shape[0], len(self.channels()), *dexp_shape[1:])
        
        group = zarr.group(zarr.NestedDirectoryStore(path))
        ome_array = group.create_dataset('0', shape=ome_zarr_shape, dtype=dtype,
                                         chunks=chunks)

        for t in range(self.nb_timepoints(ch)):
            aprint(f'Converting time point {t} ...', end='\r')
            for c, channel in enumerate(self.channels()):
                ome_array[t, c] = self.get_stack(channel, t)
        aprint('')

        group.attrs['multiscales'] = [{
            'version': CurrentFormat().version,
            'datasets': [{'path': '0'}]
        }]

        group.attrs['omero'] = default_omero_metadata(self._path, self.channels(), dtype)



import math

import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.utils import xpArray

def yang_deskew(image: xpArray,
                depth_axis: int,
                lateral_axis: int,
                flip_depth_axis: bool,
                dx: float,
                dz: float,
                angle: float,
                camera_orientation: int = 0,
                num_split: int = 4,
                internal_dtype=None
                ):
    """ 'Yang' Deskewing as done in Yang et al. 2019 ( https://www.biorxiv.org/content/10.1101/2020.09.22.309229v2 )

    Parameters
    ----------
    image  : input image (skewed 3D stack)
    depth_axis  : Depth axis.
    lateral_axis  : Lateral axis.
    flip_depth_axis   : Flips image to deskew in the opposite orientation (True for view 0 and False for view 1)
    dz     : float, scanning step (stage or galvo scanning step, not the same as the distance between the slices)
    dx     : float, pixel size of the camera
    angle  : float, incident angle of the light sheet, angle between the light sheet and the optical axis in degrees
    camera_orientation : Camera orientation correction expressed as a number of 90 deg rotations to be performed per 2D image in stack.
    num_split : number of splits to break down the data into pieces (along y, axis=2) to fit into the memory of GPU
    internal_dtype : internal dtype to perform computation

    Returns
    -------
    Deskewed image

    """

    # compute dimensionless parameters:
    xres = dx * math.cos(angle * math.pi / 180)
    resample_factor = int(round(dz / xres))
    lateral_scaling = xres / dx

    image = yang_deskew_dimensionless(image=image,
                                       depth_axis=depth_axis,
                                       lateral_axis=lateral_axis,
                                       flip_depth_axis=flip_depth_axis,
                                       resample_factor=resample_factor,
                                       lateral_scaling=lateral_scaling,
                                       camera_orientation=camera_orientation,
                                       num_split=num_split,
                                       internal_dtype=internal_dtype
                                       )

    return image

def yang_deskew_dimensionless(image: xpArray,
                depth_axis: int,
                lateral_axis: int,
                flip_depth_axis: bool,
                resample_factor: int,
                lateral_scaling: float,
                camera_orientation: int = 0,
                num_split: int = 4,
                internal_dtype=None
                ):
    """ 'Yang' Deskewing as done in Yang et al. 2019 ( https://www.biorxiv.org/content/10.1101/2020.09.22.309229v2 )

    Parameters
    ----------
    image  : input image (skewed 3D stack)
    depth_axis  : Depth axis.
    lateral_axis  : Lateral axis.
    flip_depth_axis   : Flips image to deskew in the opposite orientation (True for view 0 and False for view 1)
    resample_factor : Resampling factor
    lateral_scaling : Lateral scaling
    camera_orientation : Camera orientation correction expressed as a number of 90 deg rotations to be performed per 2D image in stack.
    num_split : number of splits to break down the data into pieces (along y, axis=2) to fit into the memory of GPU
    internal_dtype : internal dtype to perform computation

    Returns
    -------
    Deskewed image

    """
    # we don't want to move the image to the backend just now,
    # as it might be a very large image and we can actually defer moving it to the backend as
    # after splitting...
    xp = Backend.get_xp_module(image)
    sp = Backend.get_sp_module(image)

    # We save the original dtype:
    original_dtype = image.dtype

    # Default internal dtype is the same as the input image:
    if internal_dtype is None:
        internal_dtype = image.dtype

    # Numpy backend does not support float16:
    if type(Backend.current()) is NumpyBackend and internal_dtype == xp.float16:
        internal_dtype = xp.float32

    # First we compute the permutation that will reorder the axis so that the depth and lateral axis are the first axis in the image:
    permutation = (depth_axis, lateral_axis) + tuple(
        axis
        for axis in range(image.ndim)
        if axis not in [depth_axis, lateral_axis]
    )
    permutation = numpy.asarray(permutation)
    inverse_permutation = numpy.argsort(permutation)

    # We apply the permutation:
    image = xp.transpose(image, axes=permutation)

    if flip_depth_axis:
        image = xp.flip(image, axis=0)

    # rotate the data
    # Note: the weird 1+co mod 4 is due to the fact that Bin's original code was implemented for a different orientation...
    image = xp.rot90(image, k=(1+camera_orientation) % 4, axes=(1, 2))


    # deskew and rotate
    image = _resampling_vertical_split(image,
                                       resample_factor=resample_factor,
                                       lateral_scaling=lateral_scaling,
                                       num_split=num_split,
                                       internal_dtype=internal_dtype)

    # flip along axis x
    if flip_depth_axis:
        xp = Backend.get_xp_module(image)
        # _resampling_vertical_split transposes axis 0 with axis 2, so we flip along 2:
        image = xp.flip(image, axis=2)

    # We apply the inverse permutation:
    image = xp.transpose(image, axes=inverse_permutation)

    # Cast back to original dtype:
    image = image.astype(original_dtype, copy=False)

    return image


def _resampling_vertical_split(image,
                               resample_factor: int,
                               lateral_scaling: float,
                               num_split: int = 4,
                               internal_dtype=None
                               ):
    """ Same as resampling_vertical but splits the input image so that computation can fit in GPU memory.

    Parameters
    ----------
    image   : input image (skewed 3D stack)
    resample_factor : Resampling factor
    lateral_scaling : Lateral scaling
    internal_dtype : internal dtype to perform computation

    Returns
    -------
    Resampled image.
    Important note: for scalability reasons, the returned image is always numpy image.

    """

    if num_split == 1:
        output = _resampling_vertical(image,
                                      resample_factor=resample_factor,
                                      lateral_scaling=lateral_scaling,
                                      internal_dtype=internal_dtype)
    else:
        xp = Backend.get_xp_module(image)
        data_gpu_splits = xp.array_split(image, num_split, axis=1)
        data_cpu_splits = []
        for k in range(num_split):
            data_resampled = _resampling_vertical(data_gpu_splits[k],
                                                  resample_factor=resample_factor,
                                                  lateral_scaling=lateral_scaling,
                                                  internal_dtype=internal_dtype)
            data_cpu_splits.append(Backend.to_numpy(data_resampled, dtype=image.dtype))

        output = numpy.concatenate(data_cpu_splits, axis=1)

    return output


def _resampling_vertical(image,
                         resample_factor: int,
                         lateral_scaling: float,
                         internal_dtype=None):
    """ Resampling of the image by interpolation along vertical direction.
    Here we assume the dz is integer multiple of dx * cos(angle * pi / 180),
    one can also pre-interpolate the data within along the z' axis if this is not the case

    Parameters
    ----------
    image  : input image (skewed 3D stack)
    resample_factor :  Resampling factor
    lateral_scaling : Lateral scaling
    internal_dtype : internal dtype to perform computation

    Returns
    -------
    Resampled image


    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    # Move images to backend.
    image = Backend.to_backend(image, dtype=internal_dtype)

    (nz, ny, nx) = image.shape
    dtype = image.dtype

    nz_new, ny_new, nx_new = len(range(0, nx, resample_factor)), ny, nx + nz * resample_factor
    data_reassign = xp.zeros((nz_new, ny_new, nx_new), internal_dtype)

    for x in range(nx):
        x_start = x
        x_end = nz * resample_factor + x
        data_reassign[x // resample_factor, :, x_start:x_end:resample_factor] = image[:, :, x].T
    del image

    # rescale the data, interpolate along z
    data_rescale = sp.ndimage.zoom(data_reassign, zoom=(resample_factor, 1, 1), order=1)
    del data_reassign

    data_interp = xp.zeros((nz_new, ny_new, nx_new), dtype)

    for z in range(nz_new):
        for k in range(resample_factor):
            data_interp[z, :, k::resample_factor] = \
                data_rescale[z * resample_factor - k, :, k::resample_factor]
    del data_rescale

    # rescale the data, to have voxel the same along x an y;
    # remove the first z slice which has artifacts due to resampling
    image_final = sp.ndimage.zoom(data_interp[1:], zoom=(1, 1, lateral_scaling), order=1)

    return image_final
